import torch
from torch import nn


class RMSNorm(nn.Module):
	def __init__(self, dim, eps=1e-6):
		super().__init__()
		self.weight = nn.Parameter(torch.ones(dim))
		self.eps = eps
	
	def forward(self, hidden_states):
		input_dtype = hidden_states.dtype
		hidden_states = hidden_states.to(torch.float32)
		variance = hidden_states.pow(2).mean(-1, keepdim=True)  # 计算隐状态的均方根
		hidden_states = hidden_states * torch.rsqrt(variance + self.eps)
		# 将隐状态除以其均方根后重新缩放
		return self.weight * hidden_states.to(input_dtype)


if __name__ == '__main__':
	batch_size = 4
	sent_length = 10
	word_dim = 16
	X = torch.randn(batch_size, sent_length, word_dim)
	output = RMSNorm(word_dim)(X)
	print(output.size())



import torch
from typing import Tuple


def precompute_freqs_cis(dim: int, end: int, theta: float = 10000.0):
	freqs = 1.0 / (theta ** (torch.arange(0, dim, 2)[: (dim // 2)].float() / dim))
	t = torch.arange(end, device=freqs.device, dtype=torch.float32)
	freqs = torch.outer(t, freqs)
	freqs_cis = torch.polar(torch.ones_like(freqs), freqs)  # complex64
	return freqs_cis


def reshape_for_broadcast(freqs_cis: torch.Tensor, x: torch.Tensor):
	ndim = x.ndim
	assert 0 <= 1 < ndim
	assert freqs_cis.shape == (x.shape[1], x.shape[-1])
	shape = [d if i == 1 or i == ndim - 1 else 1 for i, d in enumerate(x.shape)]
	return freqs_cis.view(*shape)


def apply_rotary_emb(
		xq: torch.Tensor,
		xk: torch.Tensor,
		freqs_cis: torch.Tensor,
) -> Tuple[torch.Tensor, torch.Tensor]:
	xq_ = torch.view_as_complex(xq.float().reshape(*xq.shape[:-1], -1, 2))
	xk_ = torch.view_as_complex(xk.float().reshape(*xk.shape[:-1], -1, 2))
	freqs_cis = reshape_for_broadcast(freqs_cis, xq_)
	xq_out = torch.view_as_real(xq_ * freqs_cis).flatten(-2)
	xk_out = torch.view_as_real(xk_ * freqs_cis).flatten(-2)
	return xq_out.type_as(xq), xk_out.type_as(xk)


def apply_rotary_emb_self(inputs: torch.Tensor, freqs_cis: torch.Tensor) -> torch.Tensor:
	x_ = torch.view_as_complex(inputs.float().reshape(*inputs.shape[:-1], -1, 2))
	freqs_cis = reshape_for_broadcast(freqs_cis, x_)
	xq_out = torch.view_as_real(x_ * freqs_cis).flatten(-2)
	return xq_out.type_as(inputs)


if __name__ == '__main__':
	batch_size = 4
	seq_length = 10
	word_dim = 16
	X = torch.randn(batch_size, seq_length, word_dim)
	
	# freqs_cis = precompute_freqs_cis(word_dim, seq_length * 2)
	# print('freqs_cis1', freqs_cis.shape)
	# freqs_cis = freqs_cis[0: 0 + seq_length]
	# print('freqs_cis2', freqs_cis.shape)
	#
	# xq, xk = apply_rotary_emb(X, X, freqs_cis=freqs_cis)
	# print('xq', xq.shape, 'xk', xk.shape,)
	# print(xq[:3])
	# print(xk[:3])
	
	# rotary_emb = RotaryEmbedding(word_dim)
	# print('rotary_emb', rotary_emb(X))




import math
import operator
from functools import reduce
from typing import Union, Iterable

from .utils import batch_cut, cut


# --------------------------------------------------Token counter------------------------------------------------------
def token_counter(cut_corpus: Union[map, Iterable[str], Iterable[Iterable[str]]]):
    """
    统计词频
    :param cut_corpus: 分词后的语料
    :return: collections.Counter, 可以用items()方法取出[tuple(word, count)]
    """
    from collections import Counter
    if isinstance(cut_corpus, map) or (isinstance(cut_corpus, Iterable) and isinstance(cut_corpus[0], Iterable)):
        # return Counter([token for line in batch_cut for token in line])
        return Counter(reduce(operator.iconcat, cut_corpus, []))
    elif isinstance(cut_corpus, Iterable) and isinstance(cut_corpus[0], str):
        return Counter(cut_corpus)
    raise TypeError("'cut_corpus'参数类型不对")


# -----------------------------------------------------Text analysis----------------------------------------------
def show_label_category_count(labels):
    """
    :param corpus: 语料，可以是Iterable[str], Iterable[Iterable[str]], pandas的Series[str]
    :return: labels, label_count
    """
    import sys
    import pandas as pd
    import matplotlib.pyplot as plt
    if isinstance(labels, pd.Series):
        label_count = labels.value_counts(sort=False)
        label_count.plot(kind='bar')
        labels = labels.unique()
        min_count, max_count = label_count.min(), label_count.max()
    else:
        from collections import Counter
        label_count = Counter(labels)

        labels, counts = [], []
        min_count, max_count = sys.maxsize, 0
        for label, count in label_count.items():
            labels.append(label)
            counts.append(count)
            min_count = count if min_count > count else min_count
            max_count = count if max_count < count else max_count

        plt.bar(labels, counts, width=0.5)
        plt.xticks(ticks=labels, rotation=60)

    print(f'最大值: {max_count}，最小值: {min_count}，相差{max_count / min_count:.0f}倍')
    plt.show()
    return labels, label_count


def show_sentence_len_hist(corpus, lang='cn', cut_type='word', bins=50, scope: tuple = (0, 30), cut_fn=None):
    """
    :param corpus: 语料，可以是Iterable[str], Iterable[Iterable[str]], pandas的Series[str]
    :param lang:
    :param cut_type:
    :param bins:
    :param scope:
    :param cut_fn:
    """
    import pandas as pd
    import matplotlib.pyplot as plt
    if cut_fn is None and lang == 'cn' and cut_type == 'word':
        import jieba
        import logging
        jieba.setLogLevel(logging.INFO)
        cut_fn = jieba.lcut

    plt.figure(figsize=(12, 6))
    if isinstance(corpus, pd.Series):
        sent_length = corpus.map(lambda x: len(cut(x, lang, cut_type, cut_fn=cut_fn)))
        sent_length.hist(bins=bins)
    else:
        batch_cuts = batch_cut(corpus, lang, cut_type, cut_fn=cut_fn)
        length = list(map(lambda s: len(s), batch_cuts))
        if scope:
            length = [x for x in length if scope[0] <= x <= scope[1]]
        print(f'最短的句子是{min(length)}，最长为{max(length)}')
        plt.hist(length, bins=bins)
    plt.xlabel('sentence length')
    plt.ylabel('sentence (count)')
    plt.grid()
    plt.show()


def show_token_freq_plot(corpus, lang='cn', cut_type='word', scope: tuple = None, cut_fn=None):
    """
    :param corpus: 语料，可以是Iterable[str], Iterable[Iterable[str]], pandas的Series[str]
    :param lang:
    :param cut_type:
    :param scope:
    :param cut_fn:
    """
    import matplotlib.pyplot as plt
    if cut_fn is None and lang == 'cn' and cut_type == 'word':
        import jieba
        import logging
        jieba.setLogLevel(logging.INFO)
        cut_fn = jieba.lcut

    show_num = 10
    batch_cuts = batch_cut(corpus, lang, cut_type, cut_fn=cut_fn)
    counter = token_counter(batch_cuts)
    sorted_token_count = sorted(counter.items(), key=lambda kv: kv[1], reverse=True)
    print(f'出现频率最大的{show_num}个token:\n {sorted_token_count[:show_num]}')
    print(f'出现频率最小的{show_num}个token:\n {sorted_token_count[-show_num:]}')
    if scope is not None:
        sorted_token_count = sorted_token_count[scope[0]:scope[1]]

    sorted_count = list(map(lambda kv: kv[1], sorted_token_count))

    # sorted_count = sorted(counter.values(), reverse=True)
    # if scope is not None:
    #     token_count = sorted_count[scope[0]:scope[1]]
    plt.figure(figsize=(10, 6))
    plt.plot(list(map(lambda n: math.log(n), sorted_count)))
    # plt.plot(sorted_count, scalex='log', scaley='log')
    plt.xlabel('token:x')
    plt.ylabel('frequency:log(x)')
    plt.show()



import torch
from torch import nn
from torch.nn.utils.rnn import pack_padded_sequence, pad_packed_sequence
from nlpx.norm import RMSNorm


class CNNLayer(nn.Module):
	
	def __init__(self, embed_dim: int, seq_length: int = 16, out_channels: int = None, kernel_sizes=(2, 3, 4),
	             activation=None, layer_norm=False, batch_norm=False, bias=False):
		"""
		:param embed_dim: word embedding维度
		:param seq_length: 句子序列长度
		:param out_channels: CNN out_channels, default embed_dim
		:param kernel_sizes: size of each CNN kernel
		:param activation: CNN 激活函数
		:param layer_norm: 是否层正则化
		:param batch_norm: 是否层批则化
		:param bias:
		
		Examples::

        >>> input = torch.randn(2, 10, 100)
        >>> m = CNNLayer(embed_dim=100, seq_length=16)
        >>> output = m(input)
        >>> print(output.size())
        torch.Size([2, 48, 100])  # (batch_size, len(kernel_sizes) * seq_length, embed_dim)
		"""
		super().__init__()
		self.layer_norm = layer_norm
		out_channels = out_channels or embed_dim
		activation = activation or nn.SiLU(inplace=True)
		if batch_norm:
			self.convs = nn.ModuleList([
				nn.Sequential(
					nn.Conv1d(in_channels=embed_dim, out_channels=out_channels, kernel_size=kernel_size, bias=bias),
					nn.BatchNorm1d(num_features=out_channels),
					activation,
					nn.AdaptiveMaxPool1d(seq_length)
				) for kernel_size in kernel_sizes
			])
		else:
			self.convs = nn.ModuleList([
				nn.Sequential(
					nn.Conv1d(in_channels=embed_dim, out_channels=out_channels, kernel_size=kernel_size, bias=bias),
					activation,
					nn.AdaptiveMaxPool1d(seq_length)
				) for kernel_size in kernel_sizes
			])
		if layer_norm:
			self.norm = RMSNorm(out_channels)
	
	def forward(self, inputs: torch.Tensor):
		"""
		:param inputs: [(batch_size, sequence_length, embed_dim)]
		:return: [(batch_size, seq_length * len(kernel_sizes), out_channels)]
		"""
		inputs = inputs.transpose(2, 1)
		output = torch.cat([conv(inputs) for conv in self.convs], dim=-1)
		output = output.transpose(2, 1).contiguous()
		if self.layer_norm:
			output = self.norm(output)
		return output


class RNNLayer(nn.Module):
	"""
	Examples
	--------
	>>> from nlpx.text_token import Tokenizer
	>>> from nlpx.model import RNNAttention
	>>> tokenizer = Tokenizer(corpus)
	>>> rnn = RNNLayer(embed_dim)
	"""
	
	def __init__(self, embed_dim: int, hidden_size: int = 64, num_layers: int = 1, rnn=nn.GRU,
	             bidirectional=True, layer_norm=False, dropout: float = 0.0):
		"""
		:param embed_dim: RNN的input_size，word embedding维度
		:param hidden_size: RNN的hidden_size, RNN隐藏层维度
		:param num_layers: RNN的num_layers, RNN层数
		:param num_layers: RNN的num_layers, RNN层数
		:param rnn: 所用的RNN模型：GRU和LSTM，默认是GRU
		:param layer_norm：是否层正则化
		:param dropout：
		
		Examples::

        >>> input = torch.randn(2, 10, 100)
        >>> m = RNNLayer(embed_dim=100, hidden_size=16)
        >>> output = m(input)
        >>> print(output.size())
        torch.Size([2, 10, 32])  # (batch_size, sequence_length, 2 * hidden_size), bidirectional=True
		"""
		
		super().__init__()
		self.layer_norm = layer_norm
		self.rnn = rnn(input_size=embed_dim, hidden_size=hidden_size, num_layers=num_layers,
		               bidirectional=bidirectional, batch_first=True, dropout=dropout)
		if layer_norm:
			self.norm = RMSNorm((hidden_size << 1) if bidirectional else hidden_size)
	
	def forward(self, inputs: torch.Tensor, sequence_lengths: torch.IntTensor = None):
		"""
		:param inputs: [(batch_size, sequence_length, embed_dim)], sequence_length是不确定的
		:param sequence_lengths: [torch.IntTensor] 序列实际长度
		:return: [(batch_size, sequence_length, 2 * hidden_size)] when bidirectional is True
		:return: [(batch_size, sequence_length, hidden_size)] when bidirectional is False
		"""
		if sequence_lengths is not None and torch.all(sequence_lengths):
			output = pack_padded_sequence(inputs, sequence_lengths, batch_first=True, enforce_sorted=False)
			output, _ = self.rnn(output)
			output, _ = pad_packed_sequence(output, batch_first=True)
		else:
			output, _ = self.rnn(inputs)  # [(batch_size, sequence_length, 2 * hidden_size)]
		
		if self.layer_norm:
			output = self.norm(output)
		return output




import math
import torch
from torch import nn
from typing import Optional
from .layer import CNNLayer
from .rotary_emb import precompute_freqs_cis, apply_rotary_emb, apply_rotary_emb_self


class EmbeddingLayer(nn.Module):
	
	def __init__(self, embed_dim: int, vocab_size: Optional[int] = None, embedding=None,
	             padding_idx: Optional[int] = None, dropout: float = 0.):
		"""
		:param embed_dim: word embedding维度
		:param vocab_size: vocab size
		:param embedding: 输入的embedding，可以是nn.Embedding，torch.Tensor，list
		:param padding_idx:
		"""
		super().__init__()
		if embedding is None:
			assert vocab_size, 'vocab_size must be ge 0'
			self.embedding = nn.Embedding(vocab_size, embed_dim, padding_idx)
		elif isinstance(embedding, nn.Embedding):
			self.embedding = embedding
		elif torch.is_tensor(embedding):
			self.embedding = nn.Embedding.from_pretrained(embedding)
		else:
			self.embedding = nn.Embedding.from_pretrained(torch.tensor(embedding, dtype=torch.float))
		
		if dropout:
			self.embedding = nn.Sequential(
				self.embedding,
				nn.Dropout(dropout)
			)
			
	def forward(self, input_ids: torch.LongTensor):
		"""
		:param input_ids: [(batch_size, sequence_length, embed_dim)]
		:return: [(batch_size, sequence_length, embed_dim)]
		"""
		
		return self.embedding(input_ids)


class CNNEmbedding(nn.Module):
	
	def __init__(self, embed_dim: int, vocab_size: int = None, embedding=None, seq_length: int = 16,
	             out_channels: int = None, kernel_sizes=(2, 3, 4), activation=None, batch_norm=False, bias=False,
	             padding_idx: Optional[int] = None, dropout: float = 0.):
		"""
		:param embed_dim: word embedding维度
		:param vocab_size: vocab size
		:param embedding: 输入的embedding，可以是nn.Embedding，torch.Tensor，list
		:param seq_length: 句子序列长度
		:param out_channels: CNN out_channels, default embed_dim
		:param kernel_sizes: size of each CNN kernel
		:param activation: CNN 激活函数
		:param batch_norm: 是否批正则化
		:param bias:
		:param padding_idx:
		:param dropout:
		"""
		super().__init__()
		out_channels = out_channels or embed_dim
		if dropout:
			self.embedding = nn.Sequential(
				EmbeddingLayer(embed_dim, vocab_size, embedding, padding_idx),
				CNNLayer(embed_dim, seq_length, out_channels, kernel_sizes, activation, batch_norm, bias),
				nn.Dropout(dropout)
			)
		else:
			self.embedding = nn.Sequential(
				EmbeddingLayer(embed_dim, vocab_size, embedding, padding_idx),
				CNNLayer(embed_dim, seq_length, out_channels, kernel_sizes, activation, batch_norm, bias)
			)
	
	def forward(self, input_ids: torch.LongTensor):
		"""
		:param input_ids: [(batch_size, sequence_length, embed_dim)]
		:return: [(batch_size, seq_length * len(kernel_sizes), out_channels)]
		"""
		return self.embedding(input_ids)


class RotaryEmbedding(nn.Module):
	def __init__(self, embed_dim: int, max_length: int = 100000, theta: float = 10000.0):
		"""
		:param embed_dim: word embedding维度
		:param max_length: max length of sequences
		:param theta (float, optional): Scaling factor for frequency computation. Defaults to 10000.0.
		"""
		super().__init__()
		self.freqs_cis = precompute_freqs_cis(embed_dim, max_length * 2, theta)
		self.qw = nn.Linear(in_features=embed_dim, out_features=embed_dim)
		self.kw = nn.Linear(in_features=embed_dim, out_features=embed_dim)
	
	def forward(self, token_embedding: torch.Tensor, start_pos: int = 0):
		"""
		:param token_embedding:
		:param start_pos:
		:return: query, key with position embedding
		"""
		seq_length = token_embedding.shape[1]
		self.freqs_cis = self.freqs_cis.to(token_embedding.device)
		freqs_cis = self.freqs_cis[start_pos: start_pos + seq_length]
		query = self.qw(token_embedding)
		key = self.kw(token_embedding)
		return apply_rotary_emb(query, key, freqs_cis=freqs_cis)


class SelfRotaryEmbedding(nn.Module):
	def __init__(self, embed_dim: int, max_length: int = 100000, theta: float = 10000.0):
		"""
		:param embed_dim: word embedding维度
		:param max_length: max length of sequences
		:param theta (float, optional): Scaling factor for frequency computation. Defaults to 10000.0.
		"""
		super().__init__()
		self.freqs_cis = precompute_freqs_cis(embed_dim, max_length * 2, theta)
	
	def forward(self, token_embedding: torch.Tensor, start_pos: int = 0):
		"""
		:param token_embedding:
		:param start_pos:
		:return: query, key with position embedding
		"""
		seq_length = token_embedding.shape[1]
		self.freqs_cis = self.freqs_cis.to(token_embedding.device)
		freqs_cis = self.freqs_cis[start_pos: start_pos + seq_length]
		return apply_rotary_emb_self(token_embedding, freqs_cis=freqs_cis)


class Position(nn.Module):
	
	def __init__(self, embed_dim: int, max_length: int = 5000, alpha: float = 0.2):
		super().__init__()
		den = torch.exp(- torch.arange(0, embed_dim, 2) * math.log(10000) / embed_dim)
		pos = torch.arange(0, max_length).view(max_length, 1)
		pos_embedding = torch.zeros((max_length, embed_dim))
		pos_embedding[:, 0::2] = torch.sin(pos * den)
		pos_embedding[:, 1::2] = torch.cos(pos * den)
		pos_embedding = pos_embedding.unsqueeze(-2)
		
		self.alpha = alpha
		self.register_buffer('pos_embedding', pos_embedding)
	
	def forward(self, token_embedding: torch.Tensor):
		return (1 - self.alpha) * token_embedding + self.alpha * self.pos_embedding[:token_embedding.size(0), :]


class PositionalEncoding(nn.Module):
	
	def __init__(self, embed_dim: int, vocab_size: Optional[int] = None, embedding=None,
	             padding_idx: Optional[int] = None, max_length: int = 5000, alpha: float = 0.2, dropout: float = 0.):
		super().__init__()
		if dropout:
			self.embedding = nn.Sequential(
				EmbeddingLayer(embed_dim, vocab_size, embedding, padding_idx),
			    Position(embed_dim, max_length, alpha),
				nn.Dropout(dropout)
			)
		else:
			self.embedding = nn.Sequential(
				EmbeddingLayer(embed_dim, vocab_size, embedding, padding_idx),
				Position(embed_dim, max_length, alpha)
			)
		
	def forward(self, input_ids: torch.Tensor):
		return self.embedding(input_ids)



import os
import torch
from pathlib import Path
from typing import Union, List, Optional
from concurrent.futures import as_completed, ProcessPoolExecutor
from transformers import BertTokenizer, BertConfig, BertModel, AlbertConfig, AlbertModel, ErnieConfig, ErnieModel
from transformers.tokenization_utils import PaddingStrategy, TruncationStrategy, TensorType

from nlpx.text_token.utils import get_texts_max_length


class TokenizeVec:

	def __init__(self, pretrained_path: Union[str, Path], tokenizer=BertTokenizer):
		self.config = None
		self.pretrained = None
		self.tokenizer = tokenizer.from_pretrained(pretrained_path)

	@property
	def hidden_size(self):
		assert self.config is not None, "config must not be None"
		return self.config.hidden_size

	@staticmethod
	def get_max_length(texts: List[str], max_length: int = 0) -> int:
		if max_length and max_length > 0:
			return max_length
		return get_texts_max_length(texts, cut_type='char') + 2  # 开头结尾

	def encode_plus(self,
				texts: List[str],
				add_special_tokens: bool = True,
				padding: Union[bool, str, PaddingStrategy] = 'max_length',
				truncation: Union[bool, str, TruncationStrategy] = True,
				max_length: Optional[int] = None,
				stride: int = 0,
				is_split_into_words: bool = False,
				pad_to_multiple_of: Optional[int] = None,
				return_tensors: Optional[Union[str, TensorType]] = 'pt',
				return_token_type_ids: Optional[bool] = True,
				return_attention_mask: Optional[bool] = True,
				return_overflowing_tokens: bool = False,
				return_special_tokens_mask: bool = False,
				return_offsets_mapping: bool = False,
				return_length: bool = False,
				verbose: bool = True,
				cls: bool = False,
				**kwargs,
			) -> torch.FloatTensor:
		max_length = self.get_max_length(texts, max_length)
		return self._encode_plus(
								texts,
								add_special_tokens,
								padding,
								truncation,
								max_length,
								stride,
								is_split_into_words,
								pad_to_multiple_of,
								return_tensors,
								return_token_type_ids,
								return_attention_mask,
								return_overflowing_tokens,
								return_special_tokens_mask,
								return_offsets_mapping,
								return_length,
								verbose,
								cls,
								** kwargs
							)

	def batch_encode_plus(self,
						texts: List[str],
						add_special_tokens: bool = True,
						padding: Union[bool, str, PaddingStrategy] = 'max_length',
						truncation: Union[bool, str, TruncationStrategy] = True,
						max_length: Optional[int] = None,
						stride: int = 0,
						is_split_into_words: bool = False,
						pad_to_multiple_of: Optional[int] = None,
						return_tensors: Optional[Union[str, TensorType]] = 'pt',
						return_token_type_ids: Optional[bool] = True,
						return_attention_mask: Optional[bool] = True,
						return_overflowing_tokens: bool = False,
						return_special_tokens_mask: bool = False,
						return_offsets_mapping: bool = False,
						return_length: bool = False,
						verbose: bool = True,
						cls: bool = False,
						batch_size: int = 128,
						**kwargs,
					) -> torch.FloatTensor:
		length = len(texts)
		if length <= batch_size:
			return self.encode_plus(
									texts,
									add_special_tokens,
									padding,
									truncation,
									max_length,
									stride,
									is_split_into_words,
									pad_to_multiple_of,
									return_tensors,
									return_token_type_ids,
									return_attention_mask,
									return_overflowing_tokens,
									return_special_tokens_mask,
									return_offsets_mapping,
									return_length,
									verbose,
									cls,
									** kwargs
								)
		else:
			max_length = self.get_max_length(texts, max_length)
			text_list = self.split_texts(texts, batch_size)
			results = [self._encode_plus(
									text,
									add_special_tokens,
									padding,
									truncation,
									max_length,
									stride,
									is_split_into_words,
									pad_to_multiple_of,
									return_tensors,
									return_token_type_ids,
									return_attention_mask,
									return_overflowing_tokens,
									return_special_tokens_mask,
									return_offsets_mapping,
									return_length,
									verbose,
									cls,
									** kwargs
								) for text in text_list]
			return torch.concat(results, dim=0)

	def parallel_encode_plus(self,
						texts: List[str],
						add_special_tokens: bool = True,
						padding: Union[bool, str, PaddingStrategy] = 'max_length',
						truncation: Union[bool, str, TruncationStrategy] = True,
						max_length: Optional[int] = None,
						stride: int = 0,
						is_split_into_words: bool = False,
						pad_to_multiple_of: Optional[int] = None,
						return_tensors: Optional[Union[str, TensorType]] = 'pt',
						return_token_type_ids: Optional[bool] = True,
						return_attention_mask: Optional[bool] = True,
						return_overflowing_tokens: bool = False,
						return_special_tokens_mask: bool = False,
						return_offsets_mapping: bool = False,
						return_length: bool = False,
						verbose: bool = True,
						cls: bool = False,
						batch_size: int = 128,
						n_jobs: int = -1,
						**kwargs,
					) -> torch.FloatTensor:
		length = len(texts)
		if length <= batch_size:
			return self.encode_plus(
									texts,
									add_special_tokens,
									padding,
									truncation,
									max_length,
									stride,
									is_split_into_words,
									pad_to_multiple_of,
									return_tensors,
									return_token_type_ids,
									return_attention_mask,
									return_overflowing_tokens,
									return_special_tokens_mask,
									return_offsets_mapping,
									return_length,
									verbose,
									cls,
									** kwargs
								)
		else:
			max_length = self.get_max_length(texts, max_length)
			text_list = self.split_texts(texts, batch_size)
			if n_jobs <= 0:
				n_jobs = min(len(text_list), os.cpu_count() - 1)
			with ProcessPoolExecutor(n_jobs) as executor:
				results = [executor.submit(self._order_encode,
				                           i,
				                           text,
				                           add_special_tokens,
				                           padding,
				                           truncation,
				                           max_length,
				                           stride,
				                           is_split_into_words,
				                           pad_to_multiple_of,
				                           return_tensors,
				                           return_token_type_ids,
				                           return_attention_mask,
				                           return_overflowing_tokens,
				                           return_special_tokens_mask,
				                           return_offsets_mapping,
				                           return_length,
				                           verbose,
				                           cls,
				                           **kwargs
				                           ) for i, text in enumerate(text_list)]
				results = [r.result() for r in as_completed(results)]
			results.sort(key=lambda x: x[0], reverse=False)
			results = [r[1] for r in results]
			return torch.concat(results, dim=0)

	@staticmethod
	def split_texts(texts: List[str], batch_size: int):
		return [texts[i:i + batch_size] for i in range(0, len(texts), batch_size)]

	def _order_encode(self,
				order: int,
				texts: List[str],
				add_special_tokens: bool = True,
				padding: Union[bool, str, PaddingStrategy] = False,
				truncation: Union[bool, str, TruncationStrategy] = None,
				max_length: Optional[int] = None,
				stride: int = 0,
				is_split_into_words: bool = False,
				pad_to_multiple_of: Optional[int] = None,
				return_tensors: Optional[Union[str, TensorType]] = None,
				return_token_type_ids: Optional[bool] = None,
				return_attention_mask: Optional[bool] = None,
				return_overflowing_tokens: bool = False,
				return_special_tokens_mask: bool = False,
				return_offsets_mapping: bool = False,
				return_length: bool = False,
				verbose: bool = True,
				cls: bool = False,
				**kwargs,
			) -> torch.FloatTensor:
		return order, self._encode_plus(
								texts,
								add_special_tokens,
								padding,
								truncation,
								max_length,
								stride,
								is_split_into_words,
								pad_to_multiple_of,
								return_tensors,
								return_token_type_ids,
								return_attention_mask,
								return_overflowing_tokens,
								return_special_tokens_mask,
								return_offsets_mapping,
								return_length,
								verbose,
								cls,
								**kwargs
							)

	def _encode_plus(self,
				texts: List[str],
				add_special_tokens: bool = True,
				padding: Union[bool, str, PaddingStrategy] = False,
				truncation: Union[bool, str, TruncationStrategy] = None,
				max_length: Optional[int] = None,
				stride: int = 0,
				is_split_into_words: bool = False,
				pad_to_multiple_of: Optional[int] = None,
				return_tensors: Optional[Union[str, TensorType]] = None,
				return_token_type_ids: Optional[bool] = None,
				return_attention_mask: Optional[bool] = None,
				return_overflowing_tokens: bool = False,
				return_special_tokens_mask: bool = False,
				return_offsets_mapping: bool = False,
				return_length: bool = False,
				verbose: bool = True,
				cls: bool = False,
				**kwargs,
			) -> torch.FloatTensor:
		tokens = self.tokenizer.batch_encode_plus(
												texts,
												add_special_tokens,
												padding,
												truncation,
												max_length,
												stride,
												is_split_into_words,
												pad_to_multiple_of,
												return_tensors,
												return_token_type_ids,
												return_attention_mask,
												return_overflowing_tokens,
												return_special_tokens_mask,
												return_offsets_mapping,
												return_length,
												verbose,
												** kwargs
											)
		self.pretrained.eval()
		with torch.no_grad():
			output = self.pretrained(**tokens, output_hidden_states=True)
		return output.last_hidden_state[:, 0] if cls else output.last_hidden_state[:, 1:]

	def __call__(self,
				texts: List[str],
				add_special_tokens: bool = True,
				padding: Union[bool, str, PaddingStrategy] = 'max_length',
				truncation: Union[bool, str, TruncationStrategy] = True,
				max_length: Optional[int] = None,
				stride: int = 0,
				is_split_into_words: bool = False,
				pad_to_multiple_of: Optional[int] = None,
				return_tensors: Optional[Union[str, TensorType]] = 'pt',
				return_token_type_ids: Optional[bool] = True,
				return_attention_mask: Optional[bool] = True,
				return_overflowing_tokens: bool = False,
				return_special_tokens_mask: bool = False,
				return_offsets_mapping: bool = False,
				return_length: bool = False,
				verbose: bool = True,
				cls: bool = False,
				batch_size: int = 128,
				n_jobs: int = -1,
				**kwargs,
			) -> torch.FloatTensor:
		return self.parallel_encode_plus(
			texts,
			add_special_tokens,
			padding,
			truncation,
			max_length,
			stride,
			is_split_into_words,
			pad_to_multiple_of,
			return_tensors,
			return_token_type_ids,
			return_attention_mask,
			return_overflowing_tokens,
			return_special_tokens_mask,
			return_offsets_mapping,
			return_length,
			verbose,
			cls,
			batch_size,
			n_jobs,
			**kwargs)


class BertTokenizeVec(TokenizeVec):

	def __init__(self, pretrained_path: Union[str, Path]):
		super().__init__(pretrained_path)
		self.config = BertConfig.from_pretrained(pretrained_path)
		self.pretrained = BertModel.from_pretrained(pretrained_path, config=self.config)


class AlbertTokenizeVec(TokenizeVec):

	def __init__(self, pretrained_path: Union[str, Path]):
		super().__init__(pretrained_path)
		self.config = AlbertConfig.from_pretrained(pretrained_path)
		self.pretrained = AlbertModel.from_pretrained(pretrained_path, config=self.config)


class ErnieTokenizeVec(TokenizeVec):

	def __init__(self, pretrained_path: Union[str, Path]):
		super().__init__(pretrained_path)
		self.config = ErnieConfig.from_pretrained(pretrained_path)
		self.pretrained = ErnieModel.from_pretrained(pretrained_path, config=self.config)



from .utils import evaluate, cal_correct, acc_evaluate, r2_evaluate
from ._trainer import Trainer, SimpleTrainer, EvalTrainer, SimpleEvalTrainer, \
	ClassTrainer, SimpleClassTrainer, EvalClassTrainer, SimpleEvalClassTrainer, \
	RegressTrainer, SimpleRegressTrainer, EvalRegressTrainer, SimpleEvalRegressTrainer



import numpy as np
import pandas as pd
import torch
from torch import nn
from torch import optim
from pathlib import Path
from typing import Union, List, Tuple, Collection
from sklearn.model_selection import train_test_split
from torch.nn import functional as F
from torch.optim.lr_scheduler import LRScheduler
from torch.utils.data import DataLoader, Dataset, TensorDataset
from sklearn.metrics import mean_squared_error, root_mean_squared_error, r2_score

from nlpx.llm import TokenizeVec
from nlpx.dataset import TokenDataset, PaddingTokenCollator
from nlpx.text_token import BaseTokenizer, PaddingTokenizer, SimpleTokenizer, Tokenizer, TokenEmbedding
from nlpx.training.utils import convert_to_tensor, convert_data, convert_data_acc
from nlpx.training import evaluate, Trainer, SimpleTrainer, EvalTrainer, SimpleEvalTrainer, \
	acc_evaluate, ClassTrainer, SimpleClassTrainer, EvalClassTrainer, SimpleEvalClassTrainer, \
	r2_evaluate, RegressTrainer, SimpleRegressTrainer, EvalRegressTrainer, SimpleEvalRegressTrainer


class ModelWrapper:
	"""
	Examples
	--------
	>>> from nlpx.model.wrapper import ModelWrapper
	>>> model_wrapper = ModelWrapper(model)
	>>> model_wrapper.train(train_set, val_set, collate_fn)
	>>> model_wrapper.logits(X_test)
	"""
	
	def __init__(self, model_or_path: Union[nn.Module, str, Path], device: torch.device = None):
		self.device = device or torch.device('cuda:0' if torch.cuda.is_available() else 'cpu')
		if isinstance(model_or_path, nn.Module):
			model_or_path = model_or_path.to(self.device)
			self.model, self.best_model = model_or_path, model_or_path
		elif isinstance(model_or_path, (str, Path)):
			self.model = torch.load(model_or_path, map_location=device)
			self.best_model = self.model
	
	def train(self, train_set: Dataset, val_set: Dataset = None, collate_fn=None, epochs=100,
	          optimizer: Union[type, optim.Optimizer] = None, scheduler: LRScheduler = None,
	          lr=0.001, T_max: int = 0,
	          batch_size=64, eval_batch_size=128,
	          num_workers=0, num_eval_workers=0,
	          pin_memory: bool = False, pin_memory_device: str = "",
	          persistent_workers: bool = False,
	          early_stopping_rounds: int = None,
	          print_per_rounds: int = 1, checkpoint_per_rounds: int = 0, checkpoint_name='model.pt',
	          show_progress=False, eps=1e-5):
		if val_set:
			trainer = EvalTrainer(epochs, optimizer, scheduler, lr, T_max,
			                      batch_size, eval_batch_size,
			                      num_workers, num_eval_workers,
			                      pin_memory, pin_memory_device,
			                      persistent_workers,
			                      early_stopping_rounds,  # 早停，等10轮决策，评价指标不在变化，停止
			                      print_per_rounds,
			                      checkpoint_per_rounds,
			                      checkpoint_name,
			                      self.device)
			self.best_model, histories = trainer.train(self.model, train_set, val_set, collate_fn, show_progress, eps)
		else:
			trainer = Trainer(epochs, optimizer, scheduler, lr, T_max,
			                  batch_size,
			                  num_workers,
			                  pin_memory, pin_memory_device,
			                  persistent_workers,
			                  early_stopping_rounds,  # 早停，等10轮决策，评价指标不在变化，停止
			                  print_per_rounds,
			                  checkpoint_per_rounds,
			                  checkpoint_name,
			                  self.device)
			self.best_model, histories = trainer.train(self.model, train_set, collate_fn, show_progress, eps)
		return histories
	
	def logits(self, X: torch.Tensor):
		self.best_model.eval()
		with torch.no_grad():
			return self.best_model(X)
	
	def evaluate(self, val_set: Dataset, batch_size=64, num_workers=0, collate_fn=None):
		""" return loss """
		val_loader = DataLoader(dataset=val_set, batch_size=batch_size, num_workers=num_workers, collate_fn=collate_fn)
		return evaluate(self.best_model, val_loader, self.device)
	
	def save(self, best_model_path: Union[str, Path] = './best_model.pt',
	         last_model_path: Union[str, Path] = './last_model.pt', save_mode: str = 'both'):
		"""
		:param best_model_path: 
		:param last_model_path: 
		:param save_mode: "both" or "best" or "last"
		:return: 
		"""
		assert save_mode in ('both', 'best', 'last')
		if save_mode == 'both':
			torch.save(self.model, last_model_path)
			torch.save(self.best_model, best_model_path)
		elif save_mode == 'best':
			torch.save(self.best_model, best_model_path)
		elif save_mode == 'last':
			torch.save(self.model, last_model_path)
	
	def load(self, model_path: Union[str, Path] = './best_model.pt'):
		self.model = torch.load(model_path, map_location=self.device)
		self.best_model = self.model


class SimpleModelWrapper(ModelWrapper):
	"""
	Examples
	--------
	>>> from nlpx.model.wrapper import SimpleModelWrapper
	>>> model_wrapper = SimpleModelWrapper(model)
	>>> model_wrapper.train(X, y, collate_fn)
	>>> model_wrapper.logits(X_test)
	"""
	
	def __init__(self, model_or_path: Union[nn.Module, str, Path],
	             device: torch.device = None):
		super().__init__(model_or_path, device)
	
	def train(self, X: Union[torch.Tensor, np.ndarray, List], y: Union[torch.Tensor, np.ndarray, List],
	          val_data: Tuple[Union[torch.Tensor, np.ndarray, List], Union[torch.Tensor, np.ndarray, List]] = None,
	          collate_fn=None, epochs=100,
	          optimizer: Union[type, optim.Optimizer] = None, scheduler: LRScheduler = None,
	          lr=0.001, T_max: int = 0,
	          batch_size=64, eval_batch_size=128,
	          num_workers=0, num_eval_workers=0,
	          pin_memory: bool = False, pin_memory_device: str = "",
	          persistent_workers: bool = False,
	          early_stopping_rounds: int = None,
	          print_per_rounds: int = 1, checkpoint_per_rounds: int = 0, checkpoint_name='model.pt',
	          show_progress=False, eps=1e-5):
		if val_data:
			trainer = SimpleEvalTrainer(epochs, optimizer, scheduler, lr, T_max,
			                            batch_size, eval_batch_size,
			                            num_workers, num_eval_workers,
			                            pin_memory, pin_memory_device,
			                            persistent_workers,
			                            early_stopping_rounds,
			                            print_per_rounds,
			                            checkpoint_per_rounds,
			                            checkpoint_name,
			                            self.device)
			self.best_model, histories = trainer.train(self.model, X, y, val_data, collate_fn, show_progress, eps)
		else:
			trainer = SimpleTrainer(epochs, optimizer, scheduler, lr, T_max,
			                        batch_size,
			                        num_workers,
			                        pin_memory, pin_memory_device,
			                        persistent_workers,
			                        early_stopping_rounds,  # 早停，等10轮决策，评价指标不在变化，停止
			                        print_per_rounds,
			                        checkpoint_per_rounds,
			                        checkpoint_name,
			                        self.device)
			self.best_model, histories = trainer.train(self.model, X, y, collate_fn, show_progress, eps)
			
	def evaluate(self, X: Union[torch.Tensor, np.ndarray, List], y: Union[torch.LongTensor, np.ndarray, List],
	             batch_size=64, num_workers=0, collate_fn=None):
		""" return loss """
		X, y = convert_data(X, y)
		val_set = TensorDataset(X, y)
		return super().evaluate(val_set, batch_size, num_workers, collate_fn)


class ClassModelWrapper(ModelWrapper):
	"""
	Examples
	--------
	>>> from nlpx.model.wrapper import ClassModelWrapper
	>>> model_wrapper = ClassModelWrapper(model, classes=classes)
	>>> model_wrapper.train(train_set, val_set, collate_fn)
	>>> model_wrapper.predict(X_test)
	>>> model_wrapper.evaluate(test_set, collate_fn=collate_fn)
	"""
	
	def __init__(self, model_or_path: Union[nn.Module, str, Path], classes: Collection[str] = None,
	             device: torch.device = None):
		super().__init__(model_or_path, device)
		self.classes = classes
	
	def train(self, train_set: Dataset, val_set: Dataset = None, collate_fn=None, epochs=100,
	          optimizer: Union[type, optim.Optimizer] = None, scheduler: LRScheduler = None,
	          lr=0.001, T_max: int = 0,
	          batch_size=64, eval_batch_size=128,
	          num_workers=0, num_eval_workers=0,
	          pin_memory: bool = False, pin_memory_device: str = "",
	          persistent_workers: bool = False,
	          early_stopping_rounds: int = None,
	          print_per_rounds: int = 1, checkpoint_per_rounds: int = 0, checkpoint_name='model.pt',
	          show_progress=False, eps=1e-5):
		if val_set:
			trainer = EvalClassTrainer(epochs, optimizer, scheduler, lr, T_max,
			                           batch_size, eval_batch_size,
			                           num_workers, num_eval_workers,
			                           pin_memory, pin_memory_device,
			                           persistent_workers,
			                           early_stopping_rounds,  # 早停，等10轮决策，评价指标不在变化，停止
			                           print_per_rounds,
			                           checkpoint_per_rounds,
			                           checkpoint_name,
			                           self.device)
			self.best_model, histories = trainer.train(self.model, train_set, val_set, collate_fn, show_progress, eps)
		else:
			trainer = ClassTrainer(epochs, optimizer, scheduler, lr, T_max,
			                       batch_size, num_workers,
			                       pin_memory, pin_memory_device,
			                       persistent_workers,
			                       early_stopping_rounds,  # 早停，等10轮决策，评价指标不在变化，停止
			                       print_per_rounds,
			                       checkpoint_per_rounds,
			                       checkpoint_name,
			                       self.device)
			self.best_model, histories = trainer.train(self.model, train_set, collate_fn, show_progress, eps)
		return histories
	
	def predict(self, X: torch.Tensor):
		logits = self.logits(X)
		return logits.argmax(-1)
	
	def predict_classes(self, X: torch.Tensor):
		assert self.classes is not None, 'classes must be specified'
		pred = self.predict(X)
		return [self.classes[i] for i in pred.detach().numpy().ravel()]
	
	def predict_proba(self, X: torch.Tensor):
		logits = self.logits(X)
		result = F.softmax(logits, dim=1).max(1)
		return result.indices, result.values.numpy()
	
	def predict_classes_proba(self, X: torch.Tensor):
		assert self.classes is not None, 'classes must be specified'
		indices, values = self.predict_proba(X)
		return [self.classes[i] for i in indices.detach().numpy().ravel()], values
	
	def evaluate(self, val_set: Dataset, batch_size=64, num_workers=0, collate_fn=None):
		""" return accuracy """
		val_loader = DataLoader(dataset=val_set, batch_size=batch_size, num_workers=num_workers, collate_fn=collate_fn)
		_, acc = acc_evaluate(self.best_model, val_loader, self.device)
		return acc


class SimpleClassModelWrapper(ClassModelWrapper):
	"""
	Examples
	--------
	>>> from nlpx.model.wrapper import SimpleClassModelWrapper
	>>> model_wrapper = SimpleClassModelWrapper(model, classes=classes)
	>>> model_wrapper.train(X, y val_data, collate_fn)
	>>> model_wrapper.predict(X_test)
	>>> model_wrapper.evaluate(X_test, y_test)
	"""
	
	def __init__(self, model_or_path: Union[nn.Module, str, Path], classes: Collection[str] = None,
	             device: torch.device = None):
		super().__init__(model_or_path, classes, device)
	
	def train(self, X: Union[torch.Tensor, np.ndarray, List], y: Union[torch.LongTensor, np.ndarray, List],
	          val_data: Tuple[Union[torch.Tensor, np.ndarray, List], Union[torch.LongTensor, np.ndarray, List]] = None,
	          collate_fn=None, epochs=100,
	          optimizer: Union[type, optim.Optimizer] = None, scheduler: LRScheduler = None,
	          lr=0.001, T_max: int = 0,
	          batch_size=64, eval_batch_size=128,
	          num_workers=0, num_eval_workers=0,
	          pin_memory: bool = False, pin_memory_device: str = "",
	          persistent_workers: bool = False,
	          early_stopping_rounds: int = None,
	          print_per_rounds: int = 1, checkpoint_per_rounds: int = 0, checkpoint_name='model.pt',
	          show_progress=False, eps=1e-5):
		if val_data:
			trainer = SimpleEvalClassTrainer(epochs, optimizer, scheduler, lr, T_max,
			                                 batch_size, eval_batch_size,
			                                 num_workers, num_eval_workers,
			                                 pin_memory, pin_memory_device,
			                                 persistent_workers,
			                                 early_stopping_rounds,
			                                 print_per_rounds,
			                                 checkpoint_per_rounds,
			                                 checkpoint_name,
			                                 self.device)
			self.best_model, histories = trainer.train(self.model, X, y, val_data, collate_fn, show_progress, eps)
		else:
			trainer = SimpleClassTrainer(epochs, optimizer, scheduler, lr, T_max,
			                             batch_size, num_workers,
			                             pin_memory, pin_memory_device,
			                             persistent_workers,
			                             early_stopping_rounds,
			                             print_per_rounds,
			                             checkpoint_per_rounds,
			                             checkpoint_name,
			                             self.device)
			self.best_model, histories = trainer.train(self.model, X, y, collate_fn, show_progress, eps)
		return histories
	
	def evaluate(self, X: Union[torch.Tensor, np.ndarray, List], y: Union[torch.LongTensor, np.ndarray, List],
	             batch_size=64, num_workers=0, collate_fn=None):
		X, y = convert_data_acc(X, y)
		data_set = TensorDataset(X, y)
		return super().evaluate(data_set, batch_size, num_workers, collate_fn)


class SplitClassModelWrapper(SimpleClassModelWrapper):
	"""
	Examples
	--------
	>>> from nlpx.model.wrapper import SimpleClassModelWrapper
	>>> model_wrapper = SplitClassModelWrapper(model, classes=classes)
	>>> model_wrapper.train(X, y val_data, collate_fn)
	>>> model_wrapper.predict(X_test)
	>>> model_wrapper.evaluate(X_test, y_test)
	"""
	
	def __init__(self, model_or_path: Union[nn.Module, str, Path], classes: Collection[str] = None,
	             device: torch.device = None):
		super().__init__(model_or_path, classes, device)
	
	def train(self, X: Union[torch.Tensor, np.ndarray, List], y: Union[torch.LongTensor, np.ndarray, List],
	          val_size=0.2, random_state=None,
	          collate_fn=None, epochs=100,
	          optimizer: Union[type, optim.Optimizer] = None, scheduler: LRScheduler = None,
	          lr=0.001, T_max: int = 0,
	          batch_size=64, eval_batch_size=128,
	          num_workers=0, num_eval_workers=0,
	          pin_memory: bool = False, pin_memory_device: str = "",
	          persistent_workers: bool = False,
	          early_stopping_rounds: int = None,
	          print_per_rounds: int = 1, checkpoint_per_rounds: int = 0, checkpoint_name='model.pt',
	          show_progress=False, eps=1e-5):
		X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=val_size, random_state=random_state)
		val_data = (X_test, y_test)
		trainer = SimpleEvalClassTrainer(epochs, optimizer, scheduler, lr, T_max,
		                                 batch_size, eval_batch_size,
		                                 num_workers, num_eval_workers,
		                                 pin_memory, pin_memory_device,
		                                 persistent_workers,
		                                 early_stopping_rounds,
		                                 print_per_rounds,
		                                 checkpoint_per_rounds,
		                                 checkpoint_name,
		                                 self.device)
		self.best_model, histories = trainer.train(self.model, X, y, val_data, collate_fn, show_progress, eps)
		return histories


class TextModelWrapper(SimpleClassModelWrapper):
	"""
	Examples
	--------
	>>> from nlpx.model.wrapper import TextModelWrapper
	>>> model_wrapper = TextModelWrapper(model, tokenize_vec, classes=classes)
	>>> model_wrapper.train(train_texts, y_train val_data, collate_fn)
	>>> model_wrapper.predict(test_texts)
	>>> model_wrapper.evaluate(test_texts, y_test)
	"""
	
	def __init__(self, model_or_path: Union[nn.Module, str, Path],
	             tokenize_vec: Union[BaseTokenizer, TokenizeVec, TokenEmbedding],
	             classes: Collection[str] = None,
	             device: torch.device = None):
		super().__init__(model_or_path, classes, device)
		self.tokenize_vec = tokenize_vec
	
	def train(self, texts: Union[Collection[str], np.ndarray, pd.Series],
	          y: Union[torch.LongTensor, np.ndarray, List],
	          val_data: Tuple[
		          Union[Collection[str], np.ndarray, pd.Series], Union[torch.LongTensor, np.ndarray, List]] = None,
	          max_length: int = None, collate_fn=None, epochs=100,
	          optimizer: Union[type, optim.Optimizer] = None, scheduler: LRScheduler = None,
	          lr=0.001, T_max: int = 0,
	          batch_size=64, eval_batch_size=128,
	          num_workers=0, num_eval_workers=0,
	          pin_memory: bool = False, pin_memory_device: str = "",
	          persistent_workers: bool = False,
	          early_stopping_rounds: int = None,
	          print_per_rounds: int = 1, checkpoint_per_rounds: int = 0, checkpoint_name='model.pt',
	          n_jobs=-1, show_progress=False, eps=1e-5):
		X = self.get_vec(texts, max_length=max_length, n_jobs=n_jobs)
		if val_data:
			val_data = (self.get_vec(val_data[0], max_length=max_length, n_jobs=-1), val_data[1])
		trainer = SimpleEvalClassTrainer(epochs, optimizer, scheduler, lr, T_max,
		                                 batch_size, eval_batch_size,
		                                 num_workers, num_eval_workers,
		                                 pin_memory, pin_memory_device,
		                                 persistent_workers,
		                                 early_stopping_rounds,
		                                 print_per_rounds,
		                                 checkpoint_per_rounds,
		                                 checkpoint_name,
		                                 self.device)
		self.best_model, histories = trainer.train(self.model, X, y, val_data, collate_fn, show_progress, eps)
		return histories
	
	def predict(self, texts: Collection[str], max_length: int = None, n_jobs=-1):
		logits = self.logits(texts, max_length, n_jobs=n_jobs)
		return logits.argmax(-1)
	
	def predict_classes(self, texts: Collection[str], max_length: int = None, n_jobs=-1):
		assert self.classes is not None, 'classes must be specified'
		pred = self.predict(texts, max_length, n_jobs=n_jobs)
		return [self.classes[i] for i in pred.detach().numpy().ravel()]
	
	def predict_proba(self, texts: Collection[str], max_length: int = None, n_jobs=-1):
		logits = self.logits(texts, max_length, n_jobs=n_jobs)
		result = F.softmax(logits, dim=1).max(1)
		return result.indices, result.values.numpy()
	
	def predict_classes_proba(self, texts: Collection[str], max_length: int = None, n_jobs=-1):
		assert self.classes is not None, 'classes must be specified'
		indices, values = self.predict_proba(texts, max_length, n_jobs)
		return [self.classes[i] for i in indices.detach().numpy().ravel()], values
	
	def logits(self, texts: Collection[str], max_length: int = None, n_jobs=-1):
		X = self.get_vec(texts, max_length, n_jobs=n_jobs)
		return super().logits(X)
	
	def evaluate(self, texts: Union[str, Collection[str], np.ndarray, pd.Series],
	             y: Union[torch.LongTensor, np.ndarray, List], batch_size=64, num_workers=0,
	             max_length: int = None, collate_fn=None, n_jobs=-1):
		""" return accuracy """
		X = self.get_vec(texts, max_length, n_jobs=n_jobs)
		if isinstance(y, (np.ndarray, List)):
			y = torch.tensor(y, dtype=torch.long)
		return super().evaluate(X, y, batch_size=batch_size, num_workers=num_workers, collate_fn=collate_fn)
	
	def get_vec(self, texts: Union[str, Collection[str], np.ndarray, pd.Series], max_length: int, n_jobs: int):
		if isinstance(texts, str):
			texts = [texts]
		
		if isinstance(self.tokenize_vec, TokenizeVec):
			return self.tokenize_vec.parallel_encode_plus(texts, max_length=max_length, padding='max_length',
			                                              truncation=True, add_special_tokens=True,
			                                              return_token_type_ids=True, return_attention_mask=True,
			                                              return_tensors='pt', n_jobs=n_jobs)
		
		elif isinstance(self.tokenize_vec, (PaddingTokenizer, SimpleTokenizer, Tokenizer)):
			return torch.LongTensor(self.tokenize_vec.batch_encode(texts, max_length))
		
		elif isinstance(self.tokenize_vec, TokenEmbedding):
			return self.tokenize_vec(texts, max_length)
		
		raise ValueError("Invalid tokenize_vec, it must be a TokenizeVec or TokenEmbedding.")


class SplitTextModelWrapper(TextModelWrapper):
	"""
	Examples
	--------
	>>> from nlpx.model.wrapper import SplitTextModelWrapper
	>>> model_wrapper = SplitTextModelWrapper(model, tokenize_vec, classes=classes)
	>>> model_wrapper.train(texts, y, collate_fn)
	>>> model_wrapper.predict(test_texts)
	>>> model_wrapper.evaluate(test_texts, y_test)
	"""
	
	def __init__(self, model_or_path: Union[nn.Module, str, Path],
	             tokenize_vec: Union[BaseTokenizer, TokenizeVec, TokenEmbedding],
	             classes: Collection[str] = None,
	             device: torch.device = None):
		super().__init__(model_or_path, tokenize_vec, classes, device)
	
	def train(self, texts: Union[Collection[str], np.ndarray, pd.Series],
	          y: Union[torch.LongTensor, np.ndarray, List],
	          max_length: int = None, val_size=0.2, random_state=None, collate_fn=None, epochs=100,
	          optimizer: Union[type, optim.Optimizer] = None, scheduler: LRScheduler = None,
	          lr=0.001, T_max: int = 0,
	          batch_size=64, eval_batch_size=128,
	          num_workers=0, num_eval_workers=0,
	          pin_memory: bool = False, pin_memory_device: str = "",
	          persistent_workers: bool = False,
	          early_stopping_rounds: int = None,
	          print_per_rounds: int = 1, checkpoint_per_rounds: int = 0, checkpoint_name='model.pt',
	          n_jobs=-1, show_progress=False, eps=1e-5):
		X = self.get_vec(texts, max_length=max_length, n_jobs=n_jobs)
		X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=val_size, random_state=random_state)
		trainer = SimpleEvalClassTrainer(epochs, optimizer, scheduler, lr, T_max,
		                                 batch_size, eval_batch_size,
		                                 num_workers, num_eval_workers,
		                                 pin_memory, pin_memory_device,
		                                 persistent_workers,
		                                 early_stopping_rounds,
		                                 print_per_rounds,
		                                 checkpoint_per_rounds,
		                                 checkpoint_name,
		                                 self.device)
		self.best_model, histories = trainer.train(self.model, X, y, (X_test, y_test), collate_fn, show_progress, eps)
		return histories


class PaddingTextModelWrapper(ClassModelWrapper):
	"""
	Examples
	--------
	>>> from nlpx.model.wrapper import PaddingTextModelWrapper
	>>> model_wrapper = PaddingTextModelWrapper(model, tokenizer, classes=classes)
	>>> model_wrapper.train(train_texts, y_train val_data)
	>>> model_wrapper.predict(test_texts)
	>>> model_wrapper.evaluate(test_texts, y_test)
	"""
	
	def __init__(self, model_or_path: Union[nn.Module, str, Path],
	             tokenizer: Union[PaddingTokenizer, SimpleTokenizer, Tokenizer],
	             classes: Collection[str] = None,
	             device: torch.device = None):
		super().__init__(model_or_path, classes, device)
		self.tokenizer = tokenizer
	
	def train(self, texts: Union[Collection[str], np.ndarray, pd.Series],
	          y: Union[torch.LongTensor, np.ndarray, List],
	          val_data: Tuple[
		          Union[Collection[str], np.ndarray, pd.Series], Union[torch.LongTensor, np.ndarray, List]] = None,
	          max_length: int = None, epochs=100,
	          optimizer: Union[type, optim.Optimizer] = None, scheduler: LRScheduler = None,
	          lr=0.001, T_max: int = 0,
	          batch_size=64, eval_batch_size=128,
	          num_workers=0, num_eval_workers=0,
	          pin_memory: bool = False, pin_memory_device: str = "",
	          persistent_workers: bool = False,
	          early_stopping_rounds: int = None,
	          print_per_rounds: int = 1, checkpoint_per_rounds: int = 0, checkpoint_name='model.pt',
	          show_progress=False, eps=1e-5):
		X = self.tokenizer.batch_encode(texts, padding=False)
		train_set = TokenDataset(X, y)
		val_set = None
		if val_data:
			X_val = self.tokenizer.batch_encode(val_data[0], padding=False)
			val_set = TokenDataset(X_val, val_data[1])
		
		return super().train(train_set, val_set, collate_fn=PaddingTokenCollator(self.tokenizer.pad, max_length),
		              epochs=epochs, optimizer=optimizer, scheduler=scheduler, lr=lr,
		              T_max=T_max, batch_size=batch_size, num_workers=num_workers, pin_memory=pin_memory,
		              pin_memory_device=pin_memory_device, persistent_workers=persistent_workers,
		              early_stopping_rounds=early_stopping_rounds, print_per_rounds=print_per_rounds,
		              checkpoint_per_rounds=checkpoint_per_rounds, checkpoint_name=checkpoint_name,
		              show_progress=show_progress, eps=eps)
	
	def predict(self, texts: Collection[str], max_length: int = None):
		logits = self.logits(texts, max_length)
		return logits.argmax(-1)
	
	def predict_classes(self, texts: Collection[str], max_length: int = None):
		assert self.classes is not None, 'classes must be specified'
		pred = self.predict(texts, max_length)
		return [self.classes[i] for i in pred.detach().numpy().ravel()]
	
	def predict_proba(self, texts: Collection[str], max_length: int = None):
		logits = self.logits(texts, max_length)
		result = F.softmax(logits, dim=1).max(1)
		return result.indices, result.values.numpy()
	
	def predict_classes_proba(self, texts: Collection[str], max_length: int = None):
		assert self.classes is not None, 'classes must be specified'
		indices, values = self.predict_proba(texts, max_length)
		return [self.classes[i] for i in indices.detach().numpy().ravel()], values
	
	def logits(self, texts: Collection[str], max_length: int = None):
		X = self.tokenizer.batch_encode(texts, max_length)
		X = torch.tensor(X, dtype=torch.long)
		return super().logits(X)
	
	def evaluate(self, texts: Union[str, Collection[str], np.ndarray, pd.Series],
	             y: Union[torch.LongTensor, np.ndarray, List], batch_size=64, num_workers=0,
	             max_length: int = None):
		""" return accuracy """
		X = self.tokenizer.batch_encode(texts, padding=False)
		if isinstance(y, (np.ndarray, List)):
			y = torch.tensor(y, dtype=torch.long)
		val_set = TokenDataset(X, y)
		return super().evaluate(val_set, batch_size=batch_size, num_workers=num_workers,
		                        collate_fn=PaddingTokenCollator(self.tokenizer.pad, max_length))


class SplitPaddingTextModelWrapper(PaddingTextModelWrapper):
	"""
	Examples
	--------
	>>> from nlpx.model.wrapper import SplitPaddingTextModelWrapper
	>>> model_wrapper = SplitPaddingTextModelWrapper(tokenizer, classes=classes)
	>>> model_wrapper.train(model, texts, y)
	>>> model_wrapper.predict(test_texts)
	>>> model_wrapper.evaluate(test_texts, y_test)
	"""
	
	def __init__(self, model_or_path: Union[nn.Module, str, Path],
	             tokenizer: Union[PaddingTokenizer, SimpleTokenizer, Tokenizer],
	             classes: Collection[str] = None,
	             device: torch.device = None):
		super().__init__(model_or_path, tokenizer, classes, device)
	
	def train(self, texts: Union[Collection[str], np.ndarray, pd.Series],
	          y: Union[torch.LongTensor, np.ndarray, List],
	          max_length: int = None, val_size=0.2, random_state=None, epochs=100,
	          optimizer: Union[type, optim.Optimizer] = None, scheduler: LRScheduler = None,
	          lr=0.001, T_max: int = 0,
	          batch_size=64, eval_batch_size=128,
	          num_workers=0, num_eval_workers=0,
	          pin_memory: bool = False, pin_memory_device: str = "",
	          persistent_workers: bool = False,
	          early_stopping_rounds: int = None,
	          print_per_rounds: int = 1, checkpoint_per_rounds: int = 0, checkpoint_name='model.pt',
	          show_progress=False, eps=1e-5):
		X_train, X_test, y_train, y_test = train_test_split(texts, y, test_size=val_size, random_state=random_state)
		val_data = (X_test, y_test)
		return super().train(X_train, y_train, val_data, max_length, epochs, optimizer, scheduler, lr,
		              T_max, batch_size, eval_batch_size, num_workers, num_eval_workers, pin_memory, pin_memory_device,
		              persistent_workers, early_stopping_rounds, print_per_rounds, checkpoint_per_rounds,
		              checkpoint_name, show_progress, eps)


class RegressModelWrapper(ModelWrapper):
	"""
	Examples
	--------
	>>> from nlpx.model.wrapper import ClassModelWrapper
	>>> model_wrapper = RegressModelWrapper(model)
	>>> model_wrapper.train(train_set, val_set, collate_fn)
	>>> model_wrapper.predict(X_test)
	>>> model_wrapper.evaluate(test_set, collate_fn=collate_fn)
	"""
	
	def __init__(self, model_or_path: Union[nn.Module, str, Path], device: torch.device = None):
		super().__init__(model_or_path, device)
	
	def train(self, train_set: Dataset, val_set: Dataset = None, collate_fn=None, epochs=100,
	          optimizer: Union[type, optim.Optimizer] = None, scheduler: LRScheduler = None,
	          lr=0.001, T_max: int = 0,
	          batch_size=64, eval_batch_size=128,
	          num_workers=0, num_eval_workers=0,
	          pin_memory: bool = False, pin_memory_device: str = "",
	          persistent_workers: bool = False,
	          early_stopping_rounds: int = None,
	          print_per_rounds: int = 1, checkpoint_per_rounds: int = 0, checkpoint_name='model.pt',
	          show_progress=False, eps=1e-5):
		if val_set:
			trainer = EvalRegressTrainer(epochs, optimizer, scheduler, lr, T_max,
			                             batch_size, eval_batch_size,
			                             num_workers, num_eval_workers,
			                             pin_memory, pin_memory_device,
			                             persistent_workers,
			                             early_stopping_rounds,  # 早停，等10轮决策，评价指标不在变化，停止
			                             print_per_rounds,
			                             checkpoint_per_rounds,
			                             checkpoint_name,
			                             self.device)
			self.best_model, histories = trainer.train(self.model, train_set, val_set, collate_fn, show_progress, eps)
		else:
			trainer = RegressTrainer(epochs, optimizer, scheduler, lr, T_max,
			                         batch_size, num_workers,
			                         pin_memory, pin_memory_device,
			                         persistent_workers,
			                         early_stopping_rounds,  # 早停，等10轮决策，评价指标不在变化，停止
			                         print_per_rounds,
			                         checkpoint_per_rounds,
			                         checkpoint_name,
			                         self.device)
			self.best_model, histories = trainer.train(self.model, train_set, collate_fn, show_progress, eps)
		return histories
	
	def predict(self, X: torch.Tensor):
		return self.logits(X)
	
	def mse(self, X: Union[torch.Tensor, np.ndarray, list[float]], y: Union[torch.Tensor, np.ndarray, list[float]]) -> float:
		if isinstance(X, (List, np.ndarray)):
			X = convert_to_tensor(X, 2)
			
		if isinstance(y, list):
			y = np.array(y)
		elif torch.is_tensor(y):
			y = y.numpy()
			
		pred = self.predict(X).numpy()
		return mean_squared_error(y, pred)
	
	def rmse(self, X: Union[torch.Tensor, np.ndarray, list[float]], y: Union[torch.Tensor, np.ndarray, list[float]]) -> float:
		if isinstance(X, (List, np.ndarray)):
			X = convert_to_tensor(X, 2)
			
		if isinstance(y, list):
			y = np.array(y)
		elif torch.is_tensor(y):
			y = y.numpy()
			
		pred = self.predict(X).numpy()
		return root_mean_squared_error(y, pred)
	
	def evaluate(self, val_set: Dataset, batch_size=64, num_workers=0, collate_fn=None):
		""" return r2_score """
		val_loader = DataLoader(dataset=val_set, batch_size=batch_size, num_workers=num_workers, collate_fn=collate_fn)
		_, r2 = r2_evaluate(self.best_model, val_loader, self.device)
		return r2


class SimpleRegressModelWrapper(RegressModelWrapper):
	"""
	Examples
	--------
	>>> from nlpx.model.wrapper import ClassModelWrapper
	>>> model_wrapper = SimpleRegressModelWrapper(model)
	>>> model_wrapper.train(train_set, val_set, collate_fn)
	>>> model_wrapper.predict(X_test)
	>>> model_wrapper.evaluate(test_set, collate_fn=collate_fn)
	"""
	
	def __init__(self, model_or_path: Union[nn.Module, str, Path], device: torch.device = None):
		super().__init__(model_or_path, device)
	
	def train(self, X: Union[torch.Tensor, np.ndarray, List], y: Union[torch.Tensor, np.ndarray, List],
	          val_data: Tuple[Union[torch.Tensor, np.ndarray, List], Union[torch.Tensor, np.ndarray, List]] = None,
	          collate_fn=None, epochs=100,
	          optimizer: Union[type, optim.Optimizer] = None, scheduler: LRScheduler = None,
	          lr=0.001, T_max: int = 0,
	          batch_size=64, eval_batch_size=128,
	          num_workers=0, num_eval_workers=0,
	          pin_memory: bool = False, pin_memory_device: str = "",
	          persistent_workers: bool = False,
	          early_stopping_rounds: int = None,
	          print_per_rounds: int = 1, checkpoint_per_rounds: int = 0, checkpoint_name='model.pt',
	          show_progress=False, eps=1e-5):
		if val_data:
			trainer = SimpleEvalRegressTrainer(epochs, optimizer, scheduler, lr, T_max,
			                                   batch_size, eval_batch_size,
			                                   num_workers, num_eval_workers,
			                                   pin_memory, pin_memory_device,
			                                   persistent_workers,
			                                   early_stopping_rounds,
			                                   print_per_rounds,
			                                   checkpoint_per_rounds,
			                                   checkpoint_name,
			                                   self.device)
			self.best_model, histories = trainer.train(self.model, X, y, val_data, collate_fn, show_progress, eps)
		else:
			trainer = SimpleRegressTrainer(epochs, optimizer, scheduler, lr, T_max,
			                               batch_size,
			                               num_workers,
			                               pin_memory, pin_memory_device,
			                               persistent_workers,
			                               early_stopping_rounds,  # 早停，等10轮决策，评价指标不在变化，停止
			                               print_per_rounds,
			                               checkpoint_per_rounds,
			                               checkpoint_name,
			                               self.device)
			self.best_model, histories = trainer.train(self.model, X, y, collate_fn, show_progress, eps)
		return histories
			
	def evaluate(self, X: Union[torch.Tensor, np.ndarray, List], y: Union[torch.LongTensor, np.ndarray, List],
	             batch_size=64, num_workers=0, collate_fn=None):
		""" return loss """
		if isinstance(X, (List, np.ndarray)):
			X = convert_to_tensor(X, 2)
			
		if isinstance(y, list):
			y = np.array(y)
		elif torch.is_tensor(y):
			y = y.numpy()
		pred = self.predict(X)
		return r2_score(y, pred)


class SplitRegressModelWrapper(SimpleRegressModelWrapper):
	"""
	Examples
	--------
	>>> from nlpx.model.wrapper import SimpleClassModelWrapper
	>>> model_wrapper = SplitClassModelWrapper(model, classes=classes)
	>>> model_wrapper.train(X, y val_data, collate_fn)
	>>> model_wrapper.predict(X_test)
	>>> model_wrapper.evaluate(X_test, y_test)
	"""
	
	def __init__(self, model_or_path: Union[nn.Module, str, Path], device: torch.device = None):
		super().__init__(model_or_path, device)
	
	def train(self, X: Union[torch.Tensor, np.ndarray, List], y: Union[torch.LongTensor, np.ndarray, List],
	          val_size=0.2, random_state=None,
	          collate_fn=None, epochs=100,
	          optimizer: Union[type, optim.Optimizer] = None, scheduler: LRScheduler = None,
	          lr=0.001, T_max: int = 0,
	          batch_size=64, eval_batch_size=128,
	          num_workers=0, num_eval_workers=0,
	          pin_memory: bool = False, pin_memory_device: str = "",
	          persistent_workers: bool = False,
	          early_stopping_rounds: int = None,
	          print_per_rounds: int = 1, checkpoint_per_rounds: int = 0, checkpoint_name='model.pt',
	          show_progress=False, eps=1e-5):
		X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=val_size, random_state=random_state)
		val_data = (X_test, y_test)
		trainer = SimpleEvalRegressTrainer(epochs, optimizer, scheduler, lr, T_max,
		                                   batch_size, eval_batch_size,
		                                   num_workers, num_eval_workers,
		                                   pin_memory, pin_memory_device,
		                                   persistent_workers,
		                                   early_stopping_rounds,
		                                   print_per_rounds,
		                                   checkpoint_per_rounds,
		                                   checkpoint_name,
		                                   self.device)
		self.best_model, histories = trainer.train(self.model, X, y, val_data, collate_fn, show_progress, eps)
		return histories



import gc
import pickle
import torch
from torch.nn import Embedding
from pathlib import Path
from itertools import tee
from typing import Union, List, Collection

from .utils import UTF8, read_file, read_corpus_files, batch_cut, batch_pad_mask, cut, batch_pad, pad, \
	filter_stop_words, pad_mask, load_embedding
from .stats import token_counter


class BaseTokenizer:
	"""
	如果碰到未登录词会报错，不能padding，记住 min_freq = 0，否则易报错；如果没有特殊要求，请使用其子类SimpleTokenizer
	"""
	# SAVE_SEP = '\t'
	RESERVED_TOKENS = []
	
	def __init__(self, path: Union[str, Path] = None, texts: Collection[str] = None,
	             cut_texts: Union[Collection[Collection[str]], Collection[str]] = None, vocab: Collection[str] = None,
	             min_freq=0, special_tokens: Collection[str] = None, lang='cn', cut_type='word',
	             word_freq: bool = False,
	             stop_words: Collection[str] = None, keep_punctuation=False, cut_fn=None):
		"""
		:param path: 语料文件路径，可以是文件也可以是文件夹， 如：'./train.txt'. 如果是文件夹，会读取文件夹下的所有目录和文件.
		:param texts: 语料，每个元素是一句话，如：['上课时学生手机响个不停，老师一怒之下把手机摔了。', '家长拿发票让老师赔', ...]
		:param cut_texts: 语料，每个元素是是分词后的一句话，如：[['上课', '时', '学生', '手机', '响', '个', '不停'], ['家长', '拿', '发票', '让', '老师', '赔']]
		:param vocab: 词表，如：['上课', '学生', '手机', '不停', '，', '老师']
		:param min_freq: 最小词频，小于词词频的词会被忽略，默认是0，所有的词都保留
		:param special_tokens: 保留token, 如 '<pad>', '<unk>'等
		:param lang: 语言 'cn'和'en'
		:param cut_type: 分词类型，只支持'word'和‘char'两种类型
		:param word_freq: 是否统计词频
		:param stop_words: 停用词
		:param keep_punctuation: 是否保留标点符号
		:param cut_fn
		"""
		self.lang = lang
		self.cut_type = cut_type
		self.stop_words = stop_words,
		self.keep_punctuation = keep_punctuation
		if cut_fn is None and lang == 'cn' and cut_type == 'word':
			import jieba
			import logging
			jieba.setLogLevel(logging.INFO)
			self.cut_fn = jieba.lcut
		else:
			self.cut_fn = cut_fn
		
		self.special_tokens = self.RESERVED_TOKENS
		if special_tokens:
			self.special_tokens.extend([token for token in special_tokens if token not in self.RESERVED_TOKENS])
		
		if vocab:
			if stop_words:
				for word in stop_words:
					if word in vocab:
						vocab.remove(word)
			
			if cut_fn is None and lang == 'cn' and cut_type == 'word':
				for word in vocab:
					jieba.add_word(word)
			
			for token in {token for token in self.special_tokens if token in vocab}:
				vocab.remove(token)
			
			self.vocab = self.special_tokens + vocab if isinstance(vocab, List) else list(vocab)
			self.word_to_idx = {k: i for i, k in enumerate(self.vocab)}
			del vocab
			gc.collect()
		
		else:
			if path is not None and texts is None and cut_texts is None:
				texts = read_corpus_files(path)
			
			if texts is not None:
				cut_texts = batch_cut(texts, lang=lang, cut_type=cut_type, keep_punctuation=keep_punctuation,
				                      cut_fn=self.cut_fn)
				del texts
			
			if cut_texts is not None:
				if isinstance(cut_texts, Collection) and isinstance(cut_texts[0], str):
					cut_texts = batch_cut(cut_texts, lang='en', keep_punctuation=keep_punctuation)
				if stop_words:
					cut_texts = filter_stop_words(cut_texts, stop_words)
				counter = token_counter(cut_texts)
				sorted_token_freq = sorted(counter.items(), key=lambda kv: kv[1], reverse=True)
				self.vocab = self.special_tokens.copy()
				if min_freq > 0:
					filter_tokens = filter(lambda kv: kv[1] >= min_freq and kv[0] not in self.vocab,
					                       sorted_token_freq)
				else:
					filter_tokens = filter(lambda kv: kv[0] not in self.vocab, sorted_token_freq)
				
				if word_freq:
					filter_tokens, filter_tokens_copy = tee(filter_tokens, 2)
					self.word_freq = list(filter_tokens_copy)
				self.vocab += list(map(lambda kv: kv[0], filter_tokens))
				self.word_to_idx = {k: i for i, k in enumerate(self.vocab)}
				del sorted_token_freq, counter, cut_texts
				gc.collect()
			
			else:
				raise ValueError('参数file, texts, vocab不能同时为None.')
		
		self.special_token_ids = list(map(self._do_encode, self.special_tokens))
	
	@classmethod
	def from_file(cls, path: Union[str, Path], encoding=UTF8, pattern='*', func=read_file, min_freq: int = None,
	              special_tokens: Collection[str] = None, lang='cn', cut_type='word', word_freq=False,
	              stop_words: Collection[str] = None, keep_punctuation=False, cut_fn=None):
		"""
		:param path: 语料文件，可以是文件也可以是文件夹， 如：'./train.txt'. 如果是文件夹，会读取文件夹下的所有目录和文件.
		:param encoding: 编码
		:param pattern: 文件后缀，当file是文件夹的时候，会根据此后缀过滤文件
		:param func: 具体读取文件的处理函数，默认是read_file，可替换。注意：其函数签名为 function_name(path: str, encoding: str) -> texts: Collection[str]
		:param min_freq: 最小词频，小于词词频的词会被忽略，默认是0，所有的词都保留
		:param special_tokens: 保留token, 如 '<pad>', '<unk>'等
		:param lang: 语言 'cn'和'en'
		:param cut_type: 分词类型，只支持'word'和‘char'两种类型
		:param word_freq: 是否统计词频
		:param stop_words: 停用词
		:param keep_punctuation: 是否保留标点符号
		:param cut_fn
		"""
		texts = read_corpus_files(path, encoding, pattern, func)
		if min_freq is None:
			return cls(texts=texts, special_tokens=special_tokens, lang=lang, cut_type=cut_type,
			           word_freq=word_freq, stop_words=stop_words, keep_punctuation=keep_punctuation, cut_fn=cut_fn)
		return cls(texts=texts, min_freq=min_freq, special_tokens=special_tokens, lang=lang, cut_type=cut_type,
		           word_freq=word_freq, stop_words=stop_words, keep_punctuation=keep_punctuation, cut_fn=cut_fn)
	
	@classmethod
	def from_texts(cls, texts: Collection[str], min_freq: int = None, special_tokens: Collection[str] = None, lang='cn',
	               cut_type='word', word_freq=False, stop_words: Collection[str] = None, keep_punctuation=False,
	               cut_fn=None):
		"""
		:param texts: 语料，每个元素是一句话，如：['上课时学生手机响个不停，老师一怒之下把手机摔了。', '家长拿发票让老师赔', ....]
		:param min_freq: 最小词频，小于词词频的词会被忽略，默认是0，所有的词都保留
		:param special_tokens: 保留token, 如 '<pad>', '<unk>'等
		:param lang: 语言 'cn'和'en'
		:param cut_type: 分词类型，只支持'word'和‘char'两种类型
		:param word_freq: 是否统计词频
		:param stop_words: 停用词
		:param keep_punctuation: 是否保留标点符号
		:param cut_fn
		"""
		if min_freq is None:
			return cls(texts=texts, special_tokens=special_tokens, lang=lang, cut_type=cut_type,
			           word_freq=word_freq, stop_words=stop_words, keep_punctuation=keep_punctuation, cut_fn=cut_fn)
		return cls(texts=texts, min_freq=min_freq, special_tokens=special_tokens, lang=lang, cut_type=cut_type,
		           word_freq=word_freq, stop_words=stop_words, keep_punctuation=keep_punctuation, cut_fn=cut_fn)
	
	@classmethod
	def from_cut_texts(cls, cut_texts: Collection[Collection[str]], min_freq: int = None,
	                   special_tokens: Collection[str] = None, lang='cn', cut_type='word', word_freq=False,
	                   stop_words: Collection[str] = None, keep_punctuation=False, cut_fn=None):
		"""
		:param cut_texts: 分词后的语料，每个元素是一句话，如：[['上课', '时', '学生', '手机', '响', '个', '不停'], ['家长', '拿', '发票', '让', '老师', '赔']]
		:param min_freq: 最小词频，小于词词频的词会被忽略，默认是0，所有的词都保留
		:param special_tokens: 保留token, 如 '<pad>', '<unk>'等
		:param lang: 语言 'cn'和'en'
		:param cut_type: 分词类型，只支持'word'和‘char'两种类型
		:param word_freq: 是否统计词频
		:param stop_words: 停用词
		:param keep_punctuation: 是否保留标点符号
		:param cut_fn
		"""
		if min_freq is None:
			return cls(cut_texts=cut_texts, special_tokens=special_tokens, lang=lang, cut_type=cut_type,
			           word_freq=word_freq, stop_words=stop_words, keep_punctuation=keep_punctuation, cut_fn=cut_fn)
		return cls(cut_texts=cut_texts, min_freq=min_freq, special_tokens=special_tokens, lang=lang, cut_type=cut_type,
		           word_freq=word_freq, stop_words=stop_words, keep_punctuation=keep_punctuation, cut_fn=cut_fn)
	
	@classmethod
	def from_vocab(cls, vocab: Collection[str], special_tokens: Collection[str] = None, lang='cn', cut_type='word',
	               stop_words: Collection[str] = None, keep_punctuation=False, cut_fn=None):
		"""
		:param vocab: 词表，如：['上课', '学生', '手机', '不停', '，', '老师']
		:param special_tokens: 保留token, 如 '<pad>', '<unk>'等
		:param lang: 语言 'cn'和'en'
		:param cut_type: 分词类型，只支持'word'和‘char'两种类型
		:param stop_words: 停用词
		:param keep_punctuation: 是否保留标点符号
		:param cut_fn
		"""
		return cls(vocab=vocab, special_tokens=special_tokens, lang=lang, cut_type=cut_type, stop_words=stop_words,
		           keep_punctuation=keep_punctuation, cut_fn=cut_fn)
	
	@property
	def vocab_size(self):
		return len(self.vocab)
	
	@property
	def real_vocab(self):
		"""
		:return: 除去特殊字符的词表
		"""
		special_token_len = len(self.special_tokens)
		idx = special_token_len - 1
		if self.special_tokens[idx] == self.vocab[idx]:
			return self.vocab[len(self.special_tokens):]
		
		for i in range(idx):
			if self.special_tokens[i] != self.vocab[i]:
				idx = i
				break
		return self.vocab[idx:len(self.vocab) + idx - special_token_len]
	
	def encode(self, sentence: Union[str, Collection[str]], max_length: int = None, truncation=True,
	           is_split_into_words: bool = False):
		"""
		:param sentence: '上课时学生手机响个不停，老师一怒之下把手机摔了'
		:param max_length:
		:param truncation:
		:param is_split_into_words: 是否已经分词
		:return:
		"""
		tokens = self._encode(sentence, is_split_into_words)
		if truncation and max_length:
			return tokens[:max_length]
		return tokens
	
	def _encode(self, sentence: Union[str, Collection[str]], is_split_into_words):
		"""
		:param sentence: '上课时学生手机响个不停，老师一怒之下把手机摔了'
		:param is_split_into_words: 是否已经分词
		:return:
		"""
		if is_split_into_words:
			if isinstance(sentence, str):
				words = cut(sentence, lang='en', keep_punctuation=self.keep_punctuation)
			elif isinstance(sentence, Collection):
				words = sentence
		else:
			if isinstance(sentence, str):
				words = cut(sentence, self.lang, self.cut_type, keep_punctuation=self.keep_punctuation,
				            cut_fn=self.cut_fn)
			else:
				raise ValueError("当参数is_split_into_words为False时，参数'sentence'只能是str类型")
		
		if self.stop_words:
			return self.do_encode_with_stop_words(words)
		else:
			return self.do_encode(words)
	
	def batch_encode(self, sentences: Union[str, Collection[str], Collection[Collection[str]]], max_length: int = None,
	                 truncation=True, is_split_into_words: bool = False):
		"""
		:param sentences: ['上课时学生手机响个不停，老师一怒之下把手机摔了。', '家长拿发票让老师赔', ....]
		:param max_length:
		:param truncation:
		:param is_split_into_words: 是否已经分词
		:return:
		"""
		if isinstance(sentences, str) or (is_split_into_words and isinstance(sentences[0], str)):
			return self.encode(sentences, max_length, truncation, is_split_into_words)
		
		tokens = self._batch_encode(sentences, is_split_into_words)
		if truncation and max_length:
			return [x[:max_length] for x in tokens]
		return list(tokens)
	
	def _batch_encode(self, sentences: Union[Collection[str], Collection[Collection[str]]], is_split_into_words):
		"""
		:param sentences: ['上课时学生手机响个不停，老师一怒之下把手机摔了。', '家长拿发票让老师赔', ....]
		:param is_split_into_words: 是否已经分词
		:return:
		"""
		if isinstance(sentences, Collection):
			if is_split_into_words:
				if self.stop_words:
					return map(self.do_encode_with_stop_words, sentences)
				return map(self.do_encode, sentences)
			else:
				batch_cuts = batch_cut(sentences, lang=self.lang, cut_type=self.cut_type,
				                       keep_punctuation=self.keep_punctuation, cut_fn=self.cut_fn)
				if self.stop_words:
					return map(self.do_encode_with_stop_words, batch_cuts)
				return map(self.do_encode, batch_cuts)
		
		raise ValueError('参数"sentence"类型错误')
	
	def __call__(self, sentences: Union[Collection[str], Collection[Collection]], max_length: int = None,
	             truncation=True,
	             is_split_into_words: bool = False):
		return self.batch_encode(sentences, max_length, truncation, is_split_into_words)
	
	def do_encode(self, words: Union[str, Collection[str]]):
		"""
		把词转换成数字
		:param words: '学生' 或 ['学生', '手机', '老师']
		:return:
		"""
		if isinstance(words, str):
			return self._do_encode(words)
		return list(filter(lambda x: x is not None, map(self._do_encode, words)))
	
	def do_encode_with_stop_words(self, words: Union[str, Collection[str]]):
		"""
		把词转换成数字
		:param words: '学生' 或 ['学生', '手机', '老师']
		:return:
		"""
		if isinstance(words, str):
			return self._do_encode_with_stop_words(words)
		return list(filter(lambda x: x is not None, map(self._do_encode_with_stop_words, words)))
	
	def _do_encode(self, word: str):
		return self.word_to_idx[word]
	
	def _do_encode_with_stop_words(self, cut_token: str):
		return self.word_to_idx[cut_token] if cut_token not in self.stop_words else None
	
	def decode(self, tokens: Union[int, Collection[int]], return_special_tokens=False, return_sentence=False):
		"""
		:param tokens: [2, 19, 27, 3, 0, 0]
		:param return_special_tokens: 是否返回'<pad>', '<unk>', '<bos>', '<eos>'等特殊字符
		:param return_sentence: 返回的是一句话还是词序列
		:return: 由return_sentence决定，返回的是 '上课时学生手机响个不停‘, 还是 ['上课', '时', '学生', '手机', ’响个, '不停']
		"""
		if isinstance(tokens, int):
			return self.vocab[tokens]
		
		if return_sentence:
			return ''.join(self._decode(tokens, return_special_tokens))
		return list(self._decode(tokens, return_special_tokens))
	
	def batch_decode(self, batch_tokens: Collection[Collection[int]], return_special_tokens=False,
	                 return_sentence=False):
		"""
		:param batch_tokens: [[2, 19, 27, 3, 0, 0], [2, 10, 3, 0, 0, 0]]
		:param return_special_tokens: 是否返回'<pad>', '<unk>', '<bos>', '<eos>'等特殊字符
		:param return_sentence: 返回的是一句话还是词序列
		:return: 由return_sentence决定，返回的是 '上课时学生手机响个不停‘, 还是 ['上课', '时', '学生', '手机', ’响个, '不停']
		"""
		if isinstance(batch_tokens[0], int):
			return self.decode(batch_tokens, return_special_tokens, return_sentence)
		return list(map(lambda x: self.decode(x, return_special_tokens, return_sentence), batch_tokens))
	
	def add_special_tokens(self, special_tokens: Collection[str]):
		for token in special_tokens:
			if token in self.special_tokens:
				continue
			self.special_tokens.append(token)
			self.word_to_idx[token] = len(self.vocab)
			if token not in self.vocab:
				self.vocab.append(token)
		self.special_token_ids = list(map(self._do_encode, self.special_tokens))
	
	def add_words(self, words: Union[str, Collection[str]]):
		if isinstance(words, str):
			self.word_to_idx[words] = len(self.vocab)
			self.vocab.append(words)
		else:
			for word in words:
				if word in self.vocab:
					continue
				self.word_to_idx[word] = len(self.vocab)
				self.vocab.append(word)
	
	def add_stop_words(self, stop_words: Collection[str]):
		if not stop_words:
			return
		
		if self.stop_words is None:
			self.stop_words = stop_words if isinstance(self.stop_words, List) else list(stop_words)
			for word in stop_words:
				if word in self.vocab:
					self.vocab.remove(word)
		else:
			if not isinstance(self.stop_words, List):
				self.stop_words = list(self.stop_words)
			
			for word in stop_words:
				if word in self.stop_words:
					continue
				self.stop_words.append(word)
				if word in self.vocab:
					self.vocab.remove(word)
		
		self.word_to_idx = {k: i for i, k in enumerate(self.vocab)}
	
	def save(self, path: str = None):
		path = path or f'{self.__class__.__name__}.bin'
		obj = {
			'name': f'{self.__class__.__name__}',
			'lang': self.lang,
			'cut_type': self.cut_type,
			'keep_punctuation': self.keep_punctuation,
			'special_tokens': self.special_tokens,
			'stop_words': self.stop_words,
			'vocab': self.vocab
		}
		with open(path, 'wb') as f:
			pickle.dump(obj, f)
	
	@classmethod
	def load(cls, path: str = None, cut_fn=None):
		path = path or f'{cls.__name__}.bin'
		with open(path, 'rb') as f:
			obj = pickle.load(f)
		
		tokenizer_name = obj['name']
		assert cls.__name__ == tokenizer_name, f"Expected {tokenizer_name}, but found {cls.__name__}"
		return cls(vocab=obj['vocab'], special_tokens=obj['special_tokens'], lang=obj['lang'], cut_type=obj['cut_type'],
		           stop_words=obj['stop_words'], keep_punctuation=obj['keep_punctuation'], cut_fn=cut_fn)
	
	def _decode(self, tokens: Collection[int], return_special_tokens):
		if return_special_tokens:
			return map(lambda i: self.vocab[i], tokens)
		
		if self.special_tokens:
			return map(lambda i: self.vocab[i], filter(lambda i: i not in self.special_token_ids, tokens))
		
		return map(lambda i: self.vocab[i], tokens)
	
	def __len__(self):
		return self.vocab_size
	
	def __getitem__(self, index):
		return self.vocab[index]


class PaddingTokenizer(BaseTokenizer):
	"""
	如果碰到未登录词会用'<unk>'代替，能padding, '<pad>'和'<unk>'相同
	"""
	UNK = '<UNK>'
	PAD_ID = UNK_ID = 0
	RESERVED_TOKENS = [UNK]
	
	def __init__(self, path: Union[str, Path] = None, texts: Collection[str] = None,
	             cut_texts: Collection[Collection[str]] = None,
	             vocab: Collection[str] = None, min_freq=0, special_tokens: Collection[str] = None, lang='cn',
	             cut_type='word', word_freq: bool = False, stop_words: Collection[str] = None, keep_punctuation=False,
	             cut_fn=None):
		"""
		:param path: 语料文件路径，可以是文件也可以是文件夹， 如：'./train.txt'. 如果是文件夹，会读取文件夹下的所有目录和文件.
		:param texts: 语料，每个元素是一句话，如：['上课时学生手机响个不停，老师一怒之下把手机摔了。', '家长拿发票让老师赔', ...]
		:param cut_texts: 语料，每个元素是是分词后的一句话，如：[['上课', '时', '学生', '手机', '响', '个', '不停'], ['家长', '拿', '发票', '让', '老师', '赔']]
		:param vocab: 词表，如：['上课', '学生', '手机', '不停', '，', '老师']
		:param min_freq: 最小词频，小于词词频的词会被忽略，默认是0，所有的词都保留
		:param special_tokens: 保留token, 如 '<pad>', '<unk>'等
		:param lang: 语言 'cn'和'en'
		:param cut_type: 分词类型，只支持'word'和‘char'两种类型
		:param word_freq: 是否统计词频
		:param stop_words: 停用词
		:param keep_punctuation: 是否保留标点符号
		:param cut_fn
		"""
		super().__init__(path=path, texts=texts, cut_texts=cut_texts, vocab=vocab, min_freq=min_freq,
		                 special_tokens=special_tokens, lang=lang, cut_type=cut_type, word_freq=word_freq,
		                 stop_words=stop_words, keep_punctuation=keep_punctuation, cut_fn=cut_fn)
	
	def encode(self, sentence: str, max_length: int = None, truncation=True, padding=True, padding_side='right',
	           is_split_into_words: bool = False, return_sequence_length=False):
		"""
		:param sentence: '上课时学生手机响个不停，老师一怒之下把手机摔了'
		:param max_length:
		:param truncation:
		:param padding:
		:param padding_side:
		:param is_split_into_words: 是否已经分词
		:param return_sequence_length: 是否返回序列实际长度
		:return: ([id:int], sequence_length:int) if return_sequence_length is True, else [id:int]
		"""
		tokens = self._encode(sentence, is_split_into_words)
		if padding:
			return self.pad(tokens, max_length, truncation, padding_side, return_sequence_length)
		
		if truncation and max_length:
			tokens = tokens[:max_length]
		
		if return_sequence_length:
			return tokens, len(tokens)
		return tokens
	
	def batch_encode(self, sentences: Union[Collection[str], Collection[List]], max_length: int = None, truncation=True,
	                 padding=True, padding_side='right', is_split_into_words: bool = False,
	                 return_sequence_length=False):
		"""
		:param sentences: ['上课时学生手机响个不停，老师一怒之下把手机摔了。', '家长拿发票让老师赔', ....]
		:param max_length:
		:param truncation:
		:param padding:
		:param padding_side:
		:param is_split_into_words: 是否已经分词
		:param return_sequence_length: 是否返回序列实际长度
		:return:
			if return_sequence_length is True:
				([[id:int]], [sequence_length:int]) if tokens is batch, else ([id:int], sequence_length:int)
			
			if return_sequence_length is False:
				[[id:int]] if tokens is batch, else [id:int]
				
		"""
		if isinstance(sentences, str) or (is_split_into_words and isinstance(sentences[0], str)):
			return self.encode(sentences, max_length, truncation, padding, padding_side,
			                   is_split_into_words, return_sequence_length)
		
		tokens = self._batch_encode(sentences, is_split_into_words)
		if padding:
			return self.pad(tokens, max_length, truncation, padding_side, return_sequence_length)
		
		tokens = list(tokens)
		if truncation and max_length:
			tokens = [x[:max_length] for x in tokens]
		
		if return_sequence_length:
			return tokens, [len(x) for x in tokens]
		return tokens
	
	def pad(self, tokens: Union[map, Collection[int], Collection[Collection[int]]], max_length: int = None,
	        truncation=True,
	        padding_side='right', return_sequence_length=False):
		"""
		:param tokens: [2, 19, 27, 3] 或 [[2, 19, 27, 3], [2, 10, 3]]
		:param max_length:
		:param truncation:
		:param padding_side:
		:param return_sequence_length: 是否返回序列实际长度
		:return:
			if return_sequence_length is True:
				([[id:int]], [sequence_length:int]) if tokens is batch, else ([id:int], sequence_length:int)
			
			if return_sequence_length is False:
				[[id:int]] if tokens is batch, else [id:int]
				
		"""
		if isinstance(tokens, map) or isinstance(tokens[0], Collection):
			if return_sequence_length:
				return batch_pad(tokens, max_length, truncation, True, padding_side, False, False, self.PAD_ID, None,
				                 None)
			return batch_pad(tokens, max_length, truncation, True, padding_side, False, False, self.PAD_ID, None, None)[
				0]
		
		if return_sequence_length:
			return pad(tokens, max_length, truncation, True, padding_side, False, False, self.PAD_ID, None, None)
		return pad(tokens, max_length, truncation, True, padding_side, False, False, self.PAD_ID, None, None)[0]
	
	def __call__(self, sentences: Union[Collection[str], Collection[Collection]], max_length: int = None,
	             truncation=True,
	             padding=True, padding_side='right', is_split_into_words: bool = False,
	             return_sequence_length=False):
		return self.batch_encode(sentences, max_length, truncation, padding, padding_side, is_split_into_words,
		                         return_sequence_length)
	
	def _do_encode(self, cut_token: str):
		return self.word_to_idx.get(cut_token, self.UNK_ID)
	
	def _do_encode_with_stop_words(self, cut_token: str):
		return self.word_to_idx.get(cut_token, self.UNK_ID) if cut_token not in self.stop_words else None


class SimpleTokenizer(PaddingTokenizer):
	"""
	如果碰到未登录词会用'<unk>'代替，能padding, '<pad>'和'<unk>'不同
	"""
	PAD, UNK = '<PAD>', '<UNK>'
	PAD_ID, UNK_ID = 0, 1
	RESERVED_TOKENS = [PAD, UNK]
	
	def __init__(self, path: Union[str, Path] = None, texts: Collection[str] = None,
	             cut_texts: Collection[Collection[str]] = None,
	             vocab: Collection[str] = None, min_freq=10, special_tokens: Collection[str] = None, lang='cn',
	             cut_type='word', word_freq: bool = False, stop_words: Collection[str] = None, keep_punctuation=False,
	             cut_fn=None):
		"""
		:param path: 语料文件路径，可以是文件也可以是文件夹， 如：'./train.txt'. 如果是文件夹，会读取文件夹下的所有目录和文件.
		:param texts: 语料，每个元素是一句话，如：['上课时学生手机响个不停，老师一怒之下把手机摔了。', '家长拿发票让老师赔', ...]
		:param cut_texts: 语料，每个元素是是分词后的一句话，如：[['上课', '时', '学生', '手机', '响', '个', '不停'], ['家长', '拿', '发票', '让', '老师', '赔']]
		:param vocab: 词表，如：['上课', '学生', '手机', '不停', '，', '老师']
		:param min_freq: 最小词频，小于词词频的词会被忽略，默认是0，所有的词都保留
		:param special_tokens: 保留token, 如 '<pad>', '<unk>'等
		:param lang: 语言 'cn'和'en'
		:param cut_type: 分词类型，只支持'word'和‘char'两种类型
		:param word_freq: 是否统计词频
		:param stop_words: 停用词
		:param keep_punctuation: 是否保留标点符号
		:param cut_fn
		"""
		super().__init__(path=path, texts=texts, cut_texts=cut_texts, vocab=vocab, min_freq=min_freq,
		                 special_tokens=special_tokens, lang=lang, cut_type=cut_type, word_freq=word_freq,
		                 stop_words=stop_words, keep_punctuation=keep_punctuation, cut_fn=cut_fn)


class Tokenizer(PaddingTokenizer):
	"""
	如果碰到未登录词会用'<unk>'代替，能padding, 有encode_plus, batch_encode_plus, 返回 {'input_tokens': list, 'mask_tokens': list}
	"""
	PAD, UNK, BOS, EOS = '<PAD>', '<UNK>', '<BOS>', '<EOS>'
	PAD_ID, UNK_ID, BOS_ID, EOS_ID = 0, 1, 2, 3
	RESERVED_TOKENS = [PAD, UNK, BOS, EOS]
	
	def __init__(self, path: Union[str, Path] = None, texts: Collection[str] = None,
	             cut_texts: Collection[Collection[str]] = None,
	             vocab: Collection[str] = None, min_freq=10, special_tokens: Collection[str] = None, lang='cn',
	             cut_type='word', word_freq: bool = False, stop_words: Collection[str] = None, keep_punctuation=False,
	             cut_fn=None):
		"""
		:param path: 语料文件路径，可以是文件也可以是文件夹， 如：'./train.txt'. 如果是文件夹，会读取文件夹下的所有目录和文件.
		:param texts: 语料，每个元素是一句话，如：['上课时学生手机响个不停，老师一怒之下把手机摔了。', '家长拿发票让老师赔', ...]
		:param cut_texts: 语料，每个元素是是分词后的一句话，如：[['上课', '时', '学生', '手机', '响', '个', '不停'], ['家长', '拿', '发票', '让', '老师', '赔']]
		:param vocab: 词表，如：['上课', '学生', '手机', '不停', '，', '老师']
		:param min_freq: 最小词频，小于词词频的词会被忽略，默认是0，所有的词都保留
		:param special_tokens: 保留token, 如 '<pad>', '<unk>'等
		:param lang: 语言 'cn'和'en'
		:param cut_type: 分词类型，只支持'word'和‘char'两种类型
		:param word_freq: 是否统计词频
		:param stop_words: 停用词
		:param keep_punctuation: 是否保留标点符号
		:param cut_fn
		"""
		super().__init__(path=path, texts=texts, cut_texts=cut_texts, vocab=vocab, min_freq=min_freq,
		                 special_tokens=special_tokens, lang=lang, cut_type=cut_type, word_freq=word_freq,
		                 stop_words=stop_words, keep_punctuation=keep_punctuation, cut_fn=cut_fn)
	
	def encode(self, sentence: str, max_length: int = None, truncation=True, padding=True, padding_side='right',
	           bos=True, eos=True, is_split_into_words: bool = False, return_sequence_length=False):
		"""
		:param sentence: '上课时学生手机响个不停，老师一怒之下把手机摔了'
		:param max_length:
		:param truncation:
		:param padding:
		:param padding_side:
		:param bos:
		:param eos:
		:param is_split_into_words: 是否已经分词
		:param return_sequence_length: 是否返回序列实际长度
		:return: ([id:int], sequence_length:int) if return_sequence_length is True, else [id:int]
		"""
		tokens = self._encode(sentence, is_split_into_words)
		if padding or bos or eos:
			return self.pad(tokens, max_length, truncation, padding, padding_side, return_sequence_length, bos, eos)
		
		if truncation and max_length:
			tokens = tokens[:max_length]
		
		if return_sequence_length:
			return tokens, len(tokens)
		return tokens
	
	def batch_encode(self, sentences: Union[Collection[str], Collection[Collection]], max_length: int = None,
	                 truncation=True, padding=True, padding_side='right', bos=True, eos=True,
	                 is_split_into_words: bool = False, return_sequence_length=False):
		"""
		:param sentences: ['上课时学生手机响个不停，老师一怒之下把手机摔了。', '家长拿发票让老师赔', ....]
		:param max_length:
		:param truncation:
		:param padding:
		:param padding_side:
		:param bos:
		:param eos:
		:param is_split_into_words: 是否已经分词
		:param return_sequence_length: 是否返回序列实际长度
		:return:
			if return_sequence_length is True:
				([[id:int]], [sequence_length:int]) if tokens is batch, else ([id:int], sequence_length:int)
			
			if return_sequence_length is False:
				[[id:int]] if tokens is batch, else [id:int]
		"""
		if isinstance(sentences, str) or (is_split_into_words and isinstance(sentences[0], str)):
			return self.encode(sentences, max_length, truncation, padding, padding_side, bos, eos,
			                   is_split_into_words, return_sequence_length)
		
		tokens = self._batch_encode(sentences, is_split_into_words)
		if padding or bos or eos:
			return self.pad(tokens, max_length, truncation, padding, padding_side, return_sequence_length, bos, eos)
		
		tokens = list(tokens)
		if truncation and max_length:
			tokens = [x[:max_length] for x in tokens]
		
		if return_sequence_length:
			return tokens, [len(x) for x in tokens]
		return tokens
	
	def pad(self, tokens: Union[map, Collection[int], Collection[Collection[int]]], max_length: int = None,
	        truncation=True, padding=True, padding_side='right', return_sequence_length=False, bos=True, eos=True):
		"""
		:param tokens: [2, 19, 27, 3] 或 [[2, 19, 27, 3], [2, 10, 3]]
		:param max_length:
		:param truncation:
		:param padding
		:param padding_side:
		:param return_sequence_length: 是否返回序列实际长度
		:param bos:
		:param eos:
		:return:
			if return_sequence_length is True:
				([[id:int]], [sequence_length:int]) if tokens is batch, else ([id:int], sequence_length:int)
			
			if return_sequence_length is False:
				[[id:int]] if tokens is batch, else [id:int]
		"""
		if isinstance(tokens, map) or isinstance(tokens[0], Collection):
			if return_sequence_length:
				return batch_pad(tokens, max_length, truncation, padding, padding_side, bos, eos, self.PAD_ID,
				                 self.BOS_ID, self.EOS_ID)
			return batch_pad(tokens, max_length, truncation, padding, padding_side, bos, eos, self.PAD_ID,
			                 self.BOS_ID, self.EOS_ID)[0]
		if return_sequence_length:
			return pad(tokens, max_length, truncation, padding, padding_side, bos, eos, self.PAD_ID, self.BOS_ID,
			           self.EOS_ID)
		return \
			pad(tokens, max_length, truncation, padding, padding_side, bos, eos, self.PAD_ID, self.BOS_ID, self.EOS_ID)[
				0]
	
	def encode_plus(self, sentence: Union[str, Collection[str], Collection[Collection]], max_length: int = None,
	                truncation=True, padding=True, padding_side='right', bos=False, eos=False,
	                return_mask=False, is_split_into_words=False):
		"""
		:param sentence: '上课时学生手机响个不停，老师一怒之下把手机摔了' 或 ['上课时学生手机响个不停，老师一怒之下把手机摔了。', '家长拿发票让老师赔', ....]
		:param max_length:
		:param truncation:
		:param padding:
		:param padding_side:
		:param bos:
		:param eos:
		:param return_mask: 是否返回 mask_tokens
		:param is_split_into_words: 是否已经分词
		:return:
		"""
		if isinstance(sentence, str):
			tokens = self._encode(sentence, is_split_into_words)
			return self.pad_plus(tokens, max_length, truncation, padding, padding_side, bos, eos, return_mask)
		
		raise ValueError('参数"sentence"类型错误')
	
	def batch_encode_plus(self, sentences: Union[str, Collection[str], Collection[Collection]], max_length: int = None,
	                      truncation=True, padding=True, padding_side='right', bos=False, eos=False,
	                      return_mask=False, is_split_into_words=False):
		"""
		:param sentences: ['上课时学生手机响个不停，老师一怒之下把手机摔了。', '家长拿发票让老师赔', ....]
		:param max_length:
		:param truncation:
		:param padding:
		:param padding_side:
		:param bos:
		:param eos:
		:param return_mask: 是否返回 mask_tokens
		:param is_split_into_words: 是否已经分词
		:return:
		"""
		if isinstance(sentences, str) or (is_split_into_words and isinstance(sentences[0], str)):
			return self.encode_plus(sentences, max_length, truncation, padding, padding_side, bos, eos,
			                        return_mask, is_split_into_words)
		if isinstance(sentences, Collection):
			tokens = self._batch_encode(sentences, is_split_into_words)
			return self.pad_plus(tokens, max_length, truncation, padding, padding_side, bos, eos, return_mask)
		
		raise ValueError('参数"sentence"类型错误')
	
	def pad_plus(self, tokens: Union[map, Collection[int], Collection[Collection[int]]], max_length: int = None,
	             truncation=True, padding=True, padding_side='right', bos=False, eos=False, return_mask=False):
		"""
		:param tokens: [2, 19, 27, 3] 或 [[2, 19, 27, 3], [2, 10, 3]]
		:param max_length:
		:param truncation:
		:param padding
		:param padding_side:
		:param bos:
		:param eos:
		:param return_mask:
		:return:
		"""
		if isinstance(tokens, map) or isinstance(tokens[0], Collection):
			if return_mask:
				return batch_pad_mask(tokens, max_length, truncation, padding, padding_side, bos, eos,
				                      self.PAD_ID, self.BOS_ID, self.EOS_ID)
			return {'input_tokens': batch_pad(tokens, max_length, truncation, padding, padding_side, bos, eos,
			                                  self.PAD_ID, self.BOS_ID, self.EOS_ID)[0]}
		
		if return_mask:
			return pad_mask(tokens, max_length, truncation, padding, padding_side, bos, eos,
			                self.PAD_ID, self.BOS_ID, self.EOS_ID)
		return {'input_tokens': pad(tokens, max_length, truncation, padding, padding_side, bos, eos,
		                            self.PAD_ID, self.BOS_ID, self.EOS_ID)[0]}
	
	def __call__(self, sentence: Union[str, Collection[str]], max_length: int = None, truncation=True, padding=True,
	             padding_side='right', bos=False, eos=False, return_mask=False,
	             is_split_into_words=False):
		return self.encode_plus(sentence, max_length, truncation, padding, padding_side, bos, eos, return_mask,
		                        is_split_into_words)


class BosTokenizer(PaddingTokenizer):
	"""
	如果碰到未登录词会用'<unk>'代替，能padding, 只有BOS没有EOS
	"""
	PAD, UNK, BOS = '<PAD>', '<UNK>', '<BOS>'
	PAD_ID, UNK_ID, BOS_ID = 0, 1, 2
	RESERVED_TOKENS = [PAD, UNK, BOS]
	
	def __init__(self, path: Union[str, Path] = None, texts: Collection[str] = None,
	             cut_texts: Collection[Collection[str]] = None,
	             vocab: Collection[str] = None, min_freq=10, special_tokens: Collection[str] = None, lang='cn',
	             cut_type='word', word_freq: bool = False, stop_words: Collection[str] = None, keep_punctuation=False,
	             cut_fn=None):
		"""
		:param path: 语料文件路径，可以是文件也可以是文件夹， 如：'./train.txt'. 如果是文件夹，会读取文件夹下的所有目录和文件.
		:param texts: 语料，每个元素是一句话，如：['上课时学生手机响个不停，老师一怒之下把手机摔了。', '家长拿发票让老师赔', ...]
		:param cut_texts: 语料，每个元素是是分词后的一句话，如：[['上课', '时', '学生', '手机', '响', '个', '不停'], ['家长', '拿', '发票', '让', '老师', '赔']]
		:param vocab: 词表，如：['上课', '学生', '手机', '不停', '，', '老师']
		:param min_freq: 最小词频，小于词词频的词会被忽略，默认是0，所有的词都保留
		:param special_tokens: 保留token, 如 '<pad>', '<unk>'等
		:param lang: 语言 'cn'和'en'
		:param cut_type: 分词类型，只支持'word'和‘char'两种类型
		:param word_freq: 是否统计词频
		:param stop_words: 停用词
		:param keep_punctuation: 是否保留标点符号
		:param cut_fn
		"""
		super().__init__(path=path, texts=texts, cut_texts=cut_texts, vocab=vocab, min_freq=min_freq,
		                 special_tokens=special_tokens, lang=lang, cut_type=cut_type, word_freq=word_freq,
		                 stop_words=stop_words, keep_punctuation=keep_punctuation, cut_fn=cut_fn)
	
	def encode(self, sentence: str, max_length: int = None, truncation=True, padding=True, padding_side='right',
	           bos=True, is_split_into_words: bool = False, return_sequence_length=False):
		"""
		:param sentence: '上课时学生手机响个不停，老师一怒之下把手机摔了'
		:param max_length:
		:param truncation:
		:param padding:
		:param padding_side:
		:param bos:
		:param is_split_into_words: 是否已经分词
		:param return_sequence_length: 是否返回序列实际长度
		:return: ([id:int], sequence_length:int) if return_sequence_length is True, else [id:int]
		"""
		tokens = self._encode(sentence, is_split_into_words)
		if padding or bos:
			return self.pad(tokens, max_length, truncation, padding, padding_side, return_sequence_length, bos)
		
		if truncation and max_length:
			tokens = tokens[:max_length]
		
		if return_sequence_length:
			return tokens, len(tokens)
		return tokens
	
	def batch_encode(self, sentences: Union[Collection[str], Collection[Collection]], max_length: int = None,
	                 truncation=True, padding=True, padding_side='right', bos=True,
	                 is_split_into_words: bool = False, return_sequence_length=False):
		"""
		:param sentences: ['上课时学生手机响个不停，老师一怒之下把手机摔了。', '家长拿发票让老师赔', ....]
		:param max_length:
		:param truncation:
		:param padding:
		:param padding_side:
		:param bos:
		:param is_split_into_words: 是否已经分词
		:param return_sequence_length: 是否返回序列实际长度
		:return:
			if return_sequence_length is True:
				([[id:int]], [sequence_length:int]) if tokens is batch, else ([id:int], sequence_length:int)

			if return_sequence_length is False:
				[[id:int]] if tokens is batch, else [id:int]
		"""
		if isinstance(sentences, str) or (is_split_into_words and isinstance(sentences[0], str)):
			return self.encode(sentences, max_length, truncation, padding, padding_side, bos,
			                   is_split_into_words, return_sequence_length)
		
		tokens = self._batch_encode(sentences, is_split_into_words)
		if padding or bos:
			return self.pad(tokens, max_length, truncation, padding, padding_side, return_sequence_length, bos)
		
		tokens = list(tokens)
		if truncation and max_length:
			tokens = [x[:max_length] for x in tokens]
		
		if return_sequence_length:
			return tokens, [len(x) for x in tokens]
		return tokens
	
	def pad(self, tokens: Union[map, Collection[int], Collection[Collection[int]]], max_length: int = None,
	        truncation=True, padding=True, padding_side='right', return_sequence_length=False, bos=True):
		"""
		:param tokens: [2, 19, 27, 3] 或 [[2, 19, 27, 3], [2, 10, 3]]
		:param max_length:
		:param truncation:
		:param padding
		:param padding_side:
		:param return_sequence_length: 是否返回序列实际长度
		:param bos:
		:param eos:
		:return:
			if return_sequence_length is True:
				([[id:int]], [sequence_length:int]) if tokens is batch, else ([id:int], sequence_length:int)

			if return_sequence_length is False:
				[[id:int]] if tokens is batch, else [id:int]
		"""
		if isinstance(tokens, map) or isinstance(tokens[0], Collection):
			if return_sequence_length:
				return batch_pad(tokens, max_length, truncation, padding, padding_side, bos, False, self.PAD_ID,
				                 self.BOS_ID)
			return \
				batch_pad(tokens, max_length, truncation, padding, padding_side, bos, False, self.PAD_ID, self.BOS_ID)[
					0]
		if return_sequence_length:
			return pad(tokens, max_length, truncation, padding, padding_side, bos, False, self.PAD_ID, self.BOS_ID)
		return pad(tokens, max_length, truncation, padding, padding_side, bos, False, self.PAD_ID, self.BOS_ID)[0]


class EosTokenizer(PaddingTokenizer):
	"""
	如果碰到未登录词会用'<unk>'代替，能padding,  只有EOS没有BOS
	"""
	PAD, UNK, EOS = '<PAD>', '<UNK>', '<EOS>'
	PAD_ID, UNK_ID, EOS_ID = 0, 1, 2
	RESERVED_TOKENS = [PAD, UNK, EOS]
	
	def __init__(self, path: Union[str, Path] = None, texts: Collection[str] = None,
	             cut_texts: Collection[Collection[str]] = None,
	             vocab: Collection[str] = None, min_freq=10, special_tokens: Collection[str] = None, lang='cn',
	             cut_type='word', word_freq: bool = False, stop_words: Collection[str] = None, keep_punctuation=False,
	             cut_fn=None):
		"""
		:param path: 语料文件路径，可以是文件也可以是文件夹， 如：'./train.txt'. 如果是文件夹，会读取文件夹下的所有目录和文件.
		:param texts: 语料，每个元素是一句话，如：['上课时学生手机响个不停，老师一怒之下把手机摔了。', '家长拿发票让老师赔', ...]
		:param cut_texts: 语料，每个元素是是分词后的一句话，如：[['上课', '时', '学生', '手机', '响', '个', '不停'], ['家长', '拿', '发票', '让', '老师', '赔']]
		:param vocab: 词表，如：['上课', '学生', '手机', '不停', '，', '老师']
		:param min_freq: 最小词频，小于词词频的词会被忽略，默认是0，所有的词都保留
		:param special_tokens: 保留token, 如 '<pad>', '<unk>'等
		:param lang: 语言 'cn'和'en'
		:param cut_type: 分词类型，只支持'word'和‘char'两种类型
		:param word_freq: 是否统计词频
		:param stop_words: 停用词
		:param keep_punctuation: 是否保留标点符号
		:param cut_fn
		"""
		super().__init__(path=path, texts=texts, cut_texts=cut_texts, vocab=vocab, min_freq=min_freq,
		                 special_tokens=special_tokens, lang=lang, cut_type=cut_type, word_freq=word_freq,
		                 stop_words=stop_words, keep_punctuation=keep_punctuation, cut_fn=cut_fn)
	
	def encode(self, sentence: str, max_length: int = None, truncation=True, padding=True, padding_side='right',
	           eos=True, is_split_into_words: bool = False, return_sequence_length=False):
		"""
		:param sentence: '上课时学生手机响个不停，老师一怒之下把手机摔了'
		:param max_length:
		:param truncation:
		:param padding:
		:param padding_side:
		:param eos:
		:param is_split_into_words: 是否已经分词
		:param return_sequence_length: 是否返回序列实际长度
		:return: ([id:int], sequence_length:int) if return_sequence_length is True, else [id:int]
		"""
		tokens = self._encode(sentence, is_split_into_words)
		if padding or eos:
			return self.pad(tokens, max_length, truncation, padding, padding_side, return_sequence_length, eos)
		
		if truncation and max_length:
			tokens = tokens[:max_length]
		
		if return_sequence_length:
			return tokens, len(tokens)
		return tokens
	
	def batch_encode(self, sentences: Union[Collection[str], Collection[Collection]], max_length: int = None,
	                 truncation=True, padding=True, padding_side='right', eos=True,
	                 is_split_into_words: bool = False, return_sequence_length=False):
		"""
		:param sentences: ['上课时学生手机响个不停，老师一怒之下把手机摔了。', '家长拿发票让老师赔', ....]
		:param max_length:
		:param truncation:
		:param padding:
		:param padding_side:
		:param eos:
		:param is_split_into_words: 是否已经分词
		:param return_sequence_length: 是否返回序列实际长度
		:return:
			if return_sequence_length is True:
				([[id:int]], [sequence_length:int]) if tokens is batch, else ([id:int], sequence_length:int)

			if return_sequence_length is False:
				[[id:int]] if tokens is batch, else [id:int]
		"""
		if isinstance(sentences, str) or (is_split_into_words and isinstance(sentences[0], str)):
			return self.encode(sentences, max_length, truncation, padding, padding_side, eos,
			                   is_split_into_words, return_sequence_length)
		
		tokens = self._batch_encode(sentences, is_split_into_words)
		if padding or eos:
			return self.pad(tokens, max_length, truncation, padding, padding_side, return_sequence_length, eos)
		
		tokens = list(tokens)
		if truncation and max_length:
			tokens = [x[:max_length] for x in tokens]
		
		if return_sequence_length:
			return tokens, [len(x) for x in tokens]
		return tokens
	
	def pad(self, tokens: Union[map, Collection[int], Collection[Collection[int]]], max_length: int = None,
	        truncation=True, padding=True, padding_side='right', return_sequence_length=False, eos=True):
		"""
		:param tokens: [2, 19, 27, 3] 或 [[2, 19, 27, 3], [2, 10, 3]]
		:param max_length:
		:param truncation:
		:param padding
		:param padding_side:
		:param return_sequence_length: 是否返回序列实际长度
		:param eos:
		:return:
			if return_sequence_length is True:
				([[id:int]], [sequence_length:int]) if tokens is batch, else ([id:int], sequence_length:int)

			if return_sequence_length is False:
				[[id:int]] if tokens is batch, else [id:int]
		"""
		if isinstance(tokens, map) or isinstance(tokens[0], Collection):
			if return_sequence_length:
				return batch_pad(tokens, max_length, truncation, padding, padding_side, False, eos, self.PAD_ID,
				                 None, self.EOS_ID)
			return batch_pad(tokens, max_length, truncation, padding, padding_side, False, eos, self.PAD_ID,
			                 None, self.EOS_ID)[0]
		if return_sequence_length:
			return pad(tokens, max_length, truncation, padding, padding_side, False, eos, self.PAD_ID, None,
			           self.EOS_ID)
		return pad(tokens, max_length, truncation, padding, padding_side, False, eos, self.PAD_ID, None, self.EOS_ID)[0]


class AutoTokenizer:
	TOKENIZER_DICT = {
		"BaseTokenizer": BaseTokenizer,
		"PaddingTokenizer": PaddingTokenizer,
		"SimpleTokenizer": SimpleTokenizer,
		"Tokenizer": Tokenizer,
		"BosTokenizer": BosTokenizer,
		"EosTokenizer": EosTokenizer,
	}
	
	@staticmethod
	def load(path, cut_fn=None):
		with open(path, 'rb') as f:
			obj = pickle.load(f)
		
		tokenizer_name = obj['name']
		tokenizer = AutoTokenizer.TOKENIZER_DICT.get(tokenizer_name)
		assert tokenizer is not None, f"{tokenizer_name} is not a valid tokenizer"
		return tokenizer(vocab=obj['vocab'], special_tokens=obj['special_tokens'], lang=obj['lang'],
		                 cut_type=obj['cut_type'], stop_words=obj['stop_words'],
		                 keep_punctuation=obj['keep_punctuation'], cut_fn=cut_fn)


class TokenEmbedding:
	"""
	可以传入已经训练好的embedding文件路径，也可以embedding数据, encode返回的是 {'input_tokens': list} 或 {'input_tokens': list, 'mask_tokens': list}
	"""
	
	def __init__(self, file: Union[str, Path] = None, vocab: Collection[str] = None,
	             embedding: Collection[Collection[float]] = None,
	             special_tokens: Collection[str] = None, lang='cn', cut_type='word', func=load_embedding,
	             is_large_file=False, keep_punctuation=True, cut_fn=None):
		"""
		:param file: embedding文件路径， 如：'./sgns.weibo.word.bz2'
		:param vocab: 词表，如：['上课', '学生', '手机', '不停', '，', '老师']，与embedding必须同时传入
		:param embedding: [[0.2548, -0.6879, 0.2578],[0.2548, -0.6879, 0.2578],[0.2548, -0.6879, 0.2578]]与vocab必须同时传入
		:param special_tokens: 保留token, 如 '<pad>', '<unk>'等
		:param lang: 语言 'cn'和'en'
		:param cut_type: 分词类型，只支持'word'和‘char'两种类型
		:param func: 具体读取文件的处理函数，load_embedding，可替换。
			   注意：其函数签名为 function_name(path: str, is_large_file: bool) -> (vocab: list[str], embedding: list[list[float]])
		:param is_large_file: 是否是大文件
		:param keep_punctuation: 是否保留标点符号
		:param cut_fn
		"""
		if file:
			if isinstance(file, str):
				file = Path(file)
			assert file.is_file(), 'file必须是具体文件,不能是文件夹'
			vocab, embedding = func(file, is_large_file)
		elif not vocab or not embedding:
			raise ValueError('参数"path"为空的情况下，"vocab"和"embedding"不能为空.')
		self.tokenizer = Tokenizer(vocab=vocab, special_tokens=special_tokens, lang=lang, cut_type=cut_type,
		                           keep_punctuation=keep_punctuation, cut_fn=cut_fn)
		special_tokens = self.tokenizer.special_tokens.copy()
		special_tokens.reverse()
		self.embed_dim = len(embedding[0])
		for token in special_tokens:
			embedding = [[self.tokenizer.do_encode(token)] * self.embed_dim] + embedding
		self.embedding = Embedding.from_pretrained(torch.tensor(embedding, dtype=torch.float))
		del embedding
		gc.collect()
	
	@property
	def vocab_size(self):
		return self.tokenizer.vocab_size
	
	@property
	def real_vocab(self):
		return self.tokenizer.real_vocab
	
	def encode(self, sentence: str, max_length: int = None, truncation=True, padding=True, padding_side='right',
	           bos=False, eos=False, is_split_into_words: bool = False, return_sequence_length=False):
		"""
		:param sentence: '上课时学生手机响个不停，老师一怒之下把手机摔了'
		:param max_length:
		:param truncation:
		:param padding:
		:param padding_side:
		:param bos:
		:param eos:
		:param is_split_into_words: 是否已经分词
		:param return_sequence_length: 是否返回序列实际长度
		:return: ([id:int], sequence_length:int) if return_sequence_length is True, else [id:int]
		"""
		return self.tokenizer.encode(sentence, max_length, truncation, padding, padding_side, bos, eos,
		                             is_split_into_words, return_sequence_length)
	
	def batch_encode(self, sentences: Union[Collection[str], Collection[Collection]], max_length: int = None,
	                 truncation=True,
	                 padding=True, padding_side='right', bos=False, eos=False,
	                 is_split_into_words: bool = False, return_sequence_length=False):
		"""
		:param sentences: ['上课时学生手机响个不停，老师一怒之下把手机摔了。', '家长拿发票让老师赔', ....]
		:param max_length:
		:param truncation:
		:param padding:
		:param padding_side:
		:param bos:
		:param eos:
		:param is_split_into_words: 是否已经分词
		:param return_sequence_length: 是否返回序列实际长度
		:return:
			if return_sequence_length is True:
				([[id:int]], [sequence_length:int]) if tokens is batch, else ([id:int], sequence_length:int)
			
			if return_sequence_length is False:
				[[id:int]] if tokens is batch, else [id:int]
		"""
		return self.tokenizer.batch_encode(sentences, max_length, truncation, padding, padding_side, bos, eos,
		                                   is_split_into_words, return_sequence_length)
	
	def decode(self, tokens: Collection[int], return_special_tokens=False, return_sentence=False):
		"""
		:param tokens: [2, 19, 27, 3, 0, 0]
		:param return_special_tokens: 是否返回'<pad>', '<unk>', '<bos>', '<eos>'等特殊字符
		:param return_sentence: 返回的是一句话还是词序列
		:return: 由return_sentence决定，返回的是 '上课时学生手机响个不停‘, 还是 ['上课', '时', '学生', '手机', ’响个, '不停']
		"""
		return self.tokenizer.decode(tokens, return_special_tokens, return_sentence)
	
	def batch_decode(self, tokens: Collection[Collection[int]], return_special_tokens=False, return_sentence=False):
		"""
		:param tokens: [[2, 19, 27, 3, 0, 0], [2, 10, 3, 0, 0, 0]]
		:param return_special_tokens: 是否返回'<pad>', '<unk>', '<bos>', '<eos>'等特殊字符
		:param return_sentence: 返回的是一句话还是词序列
		:return: 由return_sentence决定，返回的是 '上课时学生手机响个不停‘, 还是 ['上课', '时', '学生', '手机', ’响个, '不停']
		"""
		return self.tokenizer.batch_decode(tokens, return_special_tokens, return_sentence)
	
	def pad(self, tokens: Union[map, Collection[int], Collection[Collection[int]]], max_length: int, truncation=True,
	        padding=True, padding_side='right', bos=False, eos=False):
		"""
		:param tokens: [2, 19, 27, 3] 或 [[2, 19, 27, 3], [2, 10, 3]]
		:param max_length:
		:param truncation:
		:param padding:
		:param padding_side:
		:param bos:
		:param eos:
		:return:
		"""
		return self.tokenizer.pad(tokens, max_length, truncation, padding, padding_side, bos, eos)
	
	def encode_plus(self, sentence: Union[str, Collection[str], Collection[Collection]], max_length: int = None,
	                truncation=True, padding=True, padding_side='right', bos=False, eos=False,
	                return_mask=False, is_split_into_words=False):
		"""
		:param sentence: '上课时学生手机响个不停，老师一怒之下把手机摔了' 或 ['上课时学生手机响个不停，老师一怒之下把手机摔了。', '家长拿发票让老师赔', ....]
		:param max_length:
		:param truncation:
		:param padding:
		:param padding_side:
		:param bos:
		:param eos:
		:param return_mask: 是否返回 mask_tokens
		:param is_split_into_words: 是否已经分词
		:return:
		"""
		return self.tokenizer.encode_plus(sentence, max_length, truncation, padding, padding_side, bos, eos,
		                                  return_mask, is_split_into_words)
	
	def batch_encode_plus(self, sentences: Union[str, Collection[str], Collection[Collection]], max_length: int = None,
	                      truncation=True, padding=True, padding_side='right', bos=False, eos=False,
	                      return_mask=False, is_split_into_words=False):
		"""
		:param sentences: ['上课时学生手机响个不停，老师一怒之下把手机摔了。', '家长拿发票让老师赔', ....]
		:param max_length:
		:param truncation:
		:param padding:
		:param padding_side:
		:param bos:
		:param eos:
		:param return_mask: 是否返回 mask_tokens
		:param is_split_into_words: 是否已经分词
		:return:
		"""
		return self.tokenizer.batch_encode_plus(sentences, max_length, truncation, padding, padding_side, bos,
		                                        eos, return_mask, is_split_into_words)
	
	def pad_plus(self, tokens: Union[map, Collection[int], Collection[Collection[int]]], max_length: int,
	             truncation=True,
	             padding=True, padding_side='right', bos=False, eos=False, return_mask=False):
		"""
		:param tokens: [2, 19, 27, 3] 或 [[2, 19, 27, 3], [2, 10, 3]]
		:param max_length:
		:param truncation:
		:param padding:
		:param padding_side:
		:param bos:
		:param eos:
		:param return_mask:
		:return:
		"""
		return self.tokenizer.pad_plus(tokens, max_length, truncation, padding, padding_side, bos, eos, return_mask)
	
	def __call__(self, sentence: Union[str, Collection[str], Collection[Collection]], max_length: int = None,
	             truncation=True,
	             padding=True, padding_side='right', bos=False, eos=False, is_split_into_words: bool = False):
		input_tokens = self.batch_encode(sentence, max_length, truncation, padding, padding_side, bos, eos,
		                                 is_split_into_words)
		return self.embedding(torch.tensor(input_tokens, dtype=torch.long))
	
	@classmethod
	def from_file(cls, file: Union[str, Path], func=load_embedding, is_large_file=False, special_tokens=[], lang='cn',
	              cut_type='word', keep_punctuation=True, cut_fn=None):
		"""
		:param file: embedding文件， 如：'./sgns.weibo.word.bz2'. 注意：必须是单一文件，不能是文件夹。
		:param func: 具体读取文件的处理函数，load_embedding，可替换。注意：其函数签名为 function_name(path: str, is_large_file: bool) -> [vocab], [[embedding]]
		:param is_large_file: 是否是大文件
		:param special_tokens: 保留token, 如 '<pad>', '<unk>'等
		:param lang: 语言 'cn'和'en'
		:param cut_type: 分词类型，只支持'word'和‘char'两种类型
		:param keep_punctuation: 是否保留标点符号
		:param cut_fn
		"""
		return cls(file=file, special_tokens=special_tokens, lang=lang, cut_type=cut_type, func=func,
		           is_large_file=is_large_file, keep_punctuation=keep_punctuation, cut_fn=cut_fn)
	
	@classmethod
	def from_vocab_embedding(cls, vocab: Collection[str], embedding: Collection[Collection[float]], large_file=False,
	                         special_tokens: Collection[str] = None, lang='cn', cut_type='word', keep_punctuation=True,
	                         cut_fn=None):
		"""
		:param vocab: 词表，如：['上课', '学生', '手机', '不停', '，', '老师']
		:param embedding: [[0.2548, -0.6879, 0.2578],[0.2548, -0.6879, 0.2578],[0.2548, -0.6879, 0.2578]]
		:param large_file: 是否是大文件
		:param special_tokens: 保留token, 如 '<pad>', '<unk>'等
		:param lang: 语言 'cn'和'en'
		:param cut_type: 分词类型，只支持'word'和‘char'两种类型
		:param keep_punctuation: 是否保留标点符号
		:param cut_fn
		"""
		return cls(vocab=vocab, embedding=embedding, large_file=large_file, special_tokens=special_tokens,
		           lang=lang, cut_type=cut_type, keep_punctuation=keep_punctuation, cut_fn=cut_fn)

# if __name__ == '__main__':
# file_name = './article.txt'
# text = read_file(file_name)

# import pandas as pd
# import matplotlib.pyplot as plt

# DATA_PATH = r'D:\Study\kkb\代码实战课\week1\toutiao-text-classify\dataset\train.csv'
# df = pd.read_csv(DATA_PATH)  # ['label']  ['sentence']

# show_label_category_count(df['label'])
# show_sentence_len_hist(df['sentence'], lang='cn', cut_type='word', bins=50, scope=(0, 30))
# show_token_freq_plot(df['sentence'], lang='cn', cut_type='word', scope=(0, 8000))

# tokenizer = Tokenizer(texts=df['sentence'].values, min_freq=10, lang='cn', cut_type='word')
# tokenizer = BaseTokenizer.from_texts(df['sentence'].values)
# tokenizer.save()

# sent = ('上课时学生手机响个不停，老师一怒之下把手机摔了，家长拿发票让老师赔，大家怎么看待这种事？', '老师一怒之下把手机摔了')
# tokenizer = BaseTokenizer.load()
# d = tokenizer.encode(sent, max_length=30, keep_punctuation=True, bos=True, eos=True)
# print(d)
# print(tokenizer.decode(d, return_special_tokens=False, return_sentence=True))
# tokenizer = Tokenizer.load()
# # print(type(tokenizer))
# # # print(len(tokenizer))
# # # print(tokenizer.vocab[:20])
# # # print(tokenizer.special_tokens)
# # # print(tokenizer.encode(['the', 'my']))
# #
# print(tokenizer.vocab[:20])
# print(sent)
# e = tokenizer(sent, max_length=30, keep_punctuation=True, bos=True, eos=True)
# print(e)
# d = tokenizer.decode(e)
# print(d)

# print(tokenizer.decode(d['input_tokens'], return_special_tokens=False, return_sentence=True))
# print(tokenizer.decode(d['input_tokens']))

# tk = [[101, 2198, 5125, 3198, 7313, 4764, 3221, 4507, 102, 7313, 4764, 3221, 4507, 102, 4764, 3221],
#       [101, 2198, 5125, 3198, 7313, 4764, 3221, 4507, 102]]
# r = tokenizer.padding(tk, max_length=15, truncation=True, padding_side='right', bos=True, eos=True)
# print(r)

# PATH = r'D:\Study\kkb\代码实战课\week1\toutiao-text-classify\dataset\sgns.weibo.word.bz2'
# tokenizer = TokenEmbedding(PATH)
# tokenizer = TokenEmbedding.from_file(PATH, is_large_file=False)
# d = tokenizer.encode(sent, max_length=30, keep_punctuation=True, bos=True, eos=True)
# print(d)
# print(tokenizer.decode(d['input_tokens'], return_special_tokens=False, return_sentence=True))
# print(tokenizer.decode(d['input_tokens']))

# tk = [[101, 2198, 5125, 3198, 7313, 4764, 3221, 4507, 102, 7313, 4764, 3221, 4507, 102, 4764, 3221],
#       [101, 2198, 5125, 3198, 7313, 4764, 3221, 4507, 102]]
# r = tokenizer.padding(tk, max_length=15, truncation=True, padding_side='right', bos=True, eos=True)
# print(r)

# print(batch_pad_mask(tk))

# tokenizer = BaseTokenizer.from_file('./', pattern='*.txt', lang='en')
# print(tokenizer.vocab)

# for f in get_files(r'D:\tmp'):
#     print(f)

# for line in read_large_file('./article.txt'):
#     print(line)
